# -*- coding: utf-8 -*-

from django.db import models, migrations
import macaddress.fields

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0003_auto_20150810_1751"),
    ]

    operations = [
        migrations.AlterField(
            model_name="interface",
            name="mac",
            field=macaddress.fields.MACAddressField(
                default=0, integer=True, null=True, db_index=True, blank=True
            ),
        ),
        migrations.AlterField(
            model_name="interface",
            name="speed",
            field=models.IntegerField(default=1000, db_index=True),
        ),
        migrations.AlterField(
            model_name="interface",
            name="type",
            field=models.IntegerField(
                default=6,
                db_index=True,
                verbose_name="Interface Type",
                choices=[
                    ("other", 1),
                    ("ethernet", 6),
                    ("loopback", 24),
                    ("tunnel", 131),
                    ("l2vlan", 135),
                    ("l3vlan", 136),
                    ("mpls", 150),
                    ("lag", 161),
                ],
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations, models

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0022_auto_20151007_1847"),
    ]

    operations = [
        migrations.AlterIndexTogether(
            name="value",
            index_together=set(
                [("resource_name", "resource_id"), ("name", "value", "resource_name")]
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations

def migrate_interface_fields(apps, schema_editor):
    """
    Migrate new Interface fields.
    """
    Interface = apps.get_model("nsot", "Interface")
    Interface_temp = apps.get_model("nsot", "Interface_temp")
    for ifc in Interface.objects.iterator():
        ifc_tmp = Interface_temp.objects.create(
            id=ifc.resource_ptr_id,
            name=ifc.name,
            description=ifc.description,
            device_id=ifc.device_id,
            parent_id=ifc.parent_id,
            type=ifc.type,
            speed=ifc.speed,
            mac_address=ifc.mac_address,
            site=ifc.site,
            _attributes_cache=ifc._attributes,
        )

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0017_move_network_data"),
    ]

    operations = [
        # Interface _attributes_cache, new_id
        migrations.RunPython(migrate_interface_fields),
    ]



from django.db import models as djmodels

from .assignment import Assignment
from .attribute import Attribute
from .change import Change
from .circuit import Circuit
from .device import Device
from .interface import Interface
from .network import Network
from .protocol import Protocol
from .protocol_type import ProtocolType
from .resource import Resource
from .site import Site
from .user import User
from .value import Value

__all__ = [
    "Assignment",
    "Attribute",
    "Change",
    "Circuit",
    "Device",
    "Interface",
    "Network",
    "Protocol",
    "ProtocolType",
    "Site",
    "User",
    "Value",
]


# Global signals
def delete_resource_values(sender, instance, **kwargs):
    """Delete values when a Resource object is deleted."""
    instance.attributes.delete()  # These are instances of Value


resource_subclasses = Resource.__subclasses__()
for model_class in resource_subclasses:
    # Value post_delete
    djmodels.signals.post_delete.connect(
        delete_resource_values,
        sender=model_class,
        dispatch_uid="value_post_delete_" + model_class.__name__,
    )



from django.contrib import admin
from django.http import JsonResponse
from django.urls import include, path, re_path
from rest_framework.renderers import JSONOpenAPIRenderer
from rest_framework.schemas import get_schema_view

from nsot import __version__

from ..api.views import NotFoundViewSet

# This is the basic API explorer for Swagger/OpenAPI 2.0
schema_view = get_schema_view(
    title="NSoT API",
    renderer_classes=[JSONOpenAPIRenderer],
)


def api_root(request):
    """JSON service info at /."""
    return JsonResponse(
        {
            "name": "Network Source of Truth (NSoT)",
            "version": __version__,
            "api": "/api/",
            "admin": "/admin/",
            "docs": "https://nsot.readthedocs.io/",
        }
    )


urlpatterns = [
    # API
    path("api/", include("nsot.api.urls")),
    # Catchall for missing endpoints
    re_path(r"^api/.*/$", NotFoundViewSet.as_view({"get": "list"})),
    # Docs
    path("schema.json", schema_view, name="swagger"),
    # Admin
    path("admin/", admin.site.urls),
    # Service info
    path("", api_root, name="index"),
]



"""
Customized base Django management command specialized for NSoT.
"""

import argparse
import logging

from django.core.management.base import BaseCommand, CommandError

__all__ = ("CommandError", "NsotCommand")


class NsotCommand(BaseCommand):
    """Base management command for NSoT that implements a custom logger."""

    # This is here to alleviate an AttributeError when getting /admin/jsi18n/
    # Ref: https://github.com/dropbox/nsot/issues/279
    leave_locale_alone = True

    def create_parser(self, prog_name, subcommand):
        """Override default parser to include default values in help."""
        parser = super().create_parser(prog_name, subcommand)

        # So that we can see default values in the help text.
        parser.formatter_class = argparse.ArgumentDefaultsHelpFormatter

        return parser

    def get_loglevel(self, verbosity, as_string=False):
        """Get the log-level."""
        if verbosity < 1:
            level_name = "notset"
        elif verbosity > 1:
            level_name = "debug"
        else:
            level_name = "info"

        if as_string:
            return level_name
        return getattr(logging, level_name.upper())

    def set_logging(self, verbosity):
        """Set the log-level."""
        log = logging.getLogger("nsot_server")

        loglevel = self.get_loglevel(verbosity)
        log.setLevel(loglevel)

        self.log = log

    def execute(self, *args, **options):
        """Setup our logging object before execution."""
        self.set_logging(options.get("verbosity"))

        super().execute(*args, **options)



import logging
from collections import OrderedDict, namedtuple

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import transaction
from django.shortcuts import get_object_or_404
from rest_framework import (
    mixins,
    permissions,
    viewsets,
)
from rest_framework import (
    status as status_codes,
)
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.views import APIView

from nsot.vendor.rest_framework_bulk import mixins as bulk_mixins

from .. import exc, models
from ..util import cidr_to_dict, qpbool
from . import auth, filters, serializers

log = logging.getLogger(__name__)


class BaseNsotViewSet(viewsets.ReadOnlyModelViewSet):
    """
    Default viewset for Nsot objects with the following defaults:

    + Objects are designed to be nested under site resources, but can also be
      top-level resources.
    """

    #: Natural key for the resource. If not defined, defaults to pk-only.
    natural_key = None

    @property
    def model_name(self):
        return self.queryset.model.__name__

    def get_natural_key_kwargs(self, filter_value):
        """
        This method should take value and return a dictionary containing the
        natural key fields used to filter results.

        This is called internally by ``self.get_object()`` if a subclass has
        defined a ``natural_key``.

        :param filter_value:
            Value to be used to filter by natural_key
        """
        return {self.natural_key: filter_value}

    def not_found(self, pk=None, site_pk=None, msg=None):
        """Standard formatting for 404 errors."""
        if msg is None:
            msg = f"No such {self.model_name} found at (site_id, id) = ({site_pk}, {pk})"
        raise exc.NotFound(msg)

    def success(self, data, status=None, headers=None):
        """
        Return a positive API response.

        :param data:
            Serialized data

        :param status:
            (Optional) HTTP status code

        :param headers:
            (Optional) Dict of extra headers
        """
        if headers is None:
            headers = self.kwargs.get("headers")

        # Default status 200 if not provided otherwise.
        if status is None:
            status = self.kwargs.get("status", status_codes.HTTP_200_OK)

        return Response(data, status=status, headers=headers)

    def list(
        self,
        request,
        site_pk=None,
        queryset=None,
        serializer_class=None,
        *args,
        **kwargs,
    ):
        """List objects optionally filtered by site.

        :param serializer_class:
            Optional serializer class override for cross-model detail routes
            (e.g. CircuitViewSet.interfaces() passing InterfaceSerializer).
        """
        if queryset is None:
            queryset = self.filter_queryset(self.get_queryset())
            # Query by site_pk if it's set (e.g. /sites/1/:foo) and make sure
            # any filtering args are passed.
            if site_pk is not None:
                queryset = queryset.filter(site=site_pk)

        if serializer_class is None:
            serializer_class = self.get_serializer_class()

        context = self.get_serializer_context()

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = serializer_class(
                page,
                many=True,
                context=context,
            )
            return super().get_paginated_response(serializer.data)

        serializer = serializer_class(
            queryset,
            many=True,
            context=context,
        )
        return self.success(serializer.data)

    def retrieve(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Retrieve a single object optionally filtered by site."""
        # If the incoming pk has changed, store it in view kwargs. This is so
        # that nested routers and detail routes work properly such as when
        # calling ``GET /api/sites/1/networks/40/parent/``.
        if "pk" in self.kwargs:
            self.kwargs["pk"] = pk
        if "site_pk" in self.kwargs:
            self.kwargs["site_pk"] = site_pk

        obj = self.get_object()
        serializer = self.get_serializer(obj, *args, **kwargs)

        return self.success(serializer.data)

    def get_object(self):
        """
        Enhanced default to support looking up objects for:

        + Natural key lookups for resource objects (e.g. Device.hostname)
        + Inject of ``site`` into filter lookup if ``site_pk`` is set.

        Currently this does NOT filter the queryset, which should not be a
        problem as we were never using .get_object() before. See the FIXME
        comments for more context.
        """
        # FIXME(jathan): Determine if we actually need .filter_queryset() here
        # queryset = self.filter_queryset(self.get_queryset())

        # FIXME(jathan): Determine if we actually need .get_queryset() here.
        # Given the way things are implemented w/ the custom .get_queryset()
        # methods for default GET filtering, we might not need to use it.
        # queryset = self.get_queryset()

        queryset = self.queryset

        # Retrieve the pk, site_pk args from the view's kwargs.
        site_pk = self.kwargs.get("site_pk")
        pk = self.kwargs.get("pk")

        # When coming from detail routes, pk might not be a string.
        if isinstance(pk, int):
            pk = str(pk)

        # Start prepping our kwargs for lookup.
        lookup_kwargs = {}
        if site_pk is not None:
            lookup_kwargs["site"] = site_pk

        # If pk is null, is a digit, or if we don't have a natural_key, lookup
        # by pk.
        if pk is None or pk.isdigit() or self.natural_key is None:
            lookup_kwargs["pk"] = pk

        # Otherwise prepare the natural_key for a single object lookup.
        else:
            natural_kwargs = self.get_natural_key_kwargs(pk)
            lookup_kwargs.update(natural_kwargs)

        # Try to get a single object.
        try:
            obj = queryset.get(**lookup_kwargs)
        except exc.ObjectDoesNotExist:
            self.not_found(pk, site_pk)
        # This may happen if:
        # - You've got multiple sites and an object w/ the same natural_key in
        #   different Sites. For example: Device 'foo-bar1' in Site 1 and
        #   Device 'foo-bar1' in Site 2.
        # - You're not using a site-specific end-point (e.g. /api/devices/ vs.
        #   /api/sites/1/devices/).
        except exc.MultipleObjectsReturned:
            raise exc.ValidationError(
                "Multiple %ss matched %s=%r. Use a site-specific endpoint "
                "or lookup by ID." % (self.model_name, self.natural_key, pk)
            )

        # May raise a permission denied.
        self.check_object_permissions(self.request, obj)

        return obj


class ChangeViewSet(BaseNsotViewSet):
    """
    Read-only API endpoint that allows Changes to be viewed.

    All Create/Update/Delete events are logged as a Change. A Change includes
    information such as the change time, user, and the full resource after
    modification. Changes are immutable and can only be removed by deleting the
    entire Site.
    """

    queryset = models.Change.objects.order_by("-change_at")
    serializer_class = serializers.ChangeSerializer
    filterset_class = filters.ChangeFilter

    @action(methods=["get"], detail=True)
    def diff(self, request, *args, **kwargs):
        return self.success(self.get_object().resource_diff)


class NsotViewSet(BaseNsotViewSet, viewsets.ModelViewSet):
    """
    Generic mutable viewset that logs all change events and includes support
    for bulk creation of objects.
    """

    def perform_create(self, serializer):
        """Support bulk create.

        :param serializer:
            Serializer instance
        """
        try:
            objects = serializer.save()
        except exc.DjangoValidationError as err:
            raise exc.ValidationError(err.message_dict)
        except exc.IntegrityError as err:
            raise exc.Conflict(str(err))
        except exc.ObjectDoesNotExist:
            raise exc.BadRequest(
                "Site with id number %s does not exist"
                % self.kwargs["site_pk"]
            )
        else:
            # This is so that we can always work w/ objects as a list
            if not isinstance(objects, list):
                objects = [objects]

        log.debug("NsotViewSet.perform_create() objects = %r", objects)
        for obj in objects:
            models.Change.objects.create(
                obj=obj, user=self.request.user, event="Create"
            )

    def get_success_headers(self, data):
        """
        Overload default to include relative request PATH.

        :param data:
            Dict of validated serializer data
        """
        # TODO(jathan): Implement hyperlinked fields in the API?
        try:
            # return {'Location': data[api_settings.URL_FIELD_NAME]}
            location = "%s%s/" % (self.request.path_info, data["id"])
            return {"Location": location}
        except (TypeError, KeyError):
            return {}

    def perform_update(self, serializer):
        """
        Overload default to handle non-serializer exceptions, and log Change
        events.

        :param serializer:
            Serializer instance
        """
        try:
            objects = serializer.save()
        except exc.DjangoValidationError as err:
            raise exc.ValidationError(err.message_dict)
        except exc.IntegrityError as err:
            raise exc.Conflict(str(err))
        else:
            # This is so that we can always work w/ objects as a list
            if not isinstance(objects, list):
                objects = [objects]

        log.debug("NsotViewSet.perform_update() objects = %r", objects)
        for obj in objects:
            models.Change.objects.create(
                obj=obj, user=self.request.user, event="Update"
            )

    def perform_destroy(self, instance):
        """
        Overload default to handle non-serializer exceptions, and log Change
        events.

        :param instance:
            Model instance to delete
        """
        log.debug("NsotViewSet.perform_destroy() obj = %r", instance)
        change = models.Change.objects.create(
            obj=instance, user=self.request.user, event="Delete"
        )

        try:
            instance.delete()
        except exc.ProtectedError as err:
            change.delete()
            raise exc.Conflict(err.args[0])
        except exc.ValidationError as err:
            change.delete()
            raise

    def bulk_destroy(self, request, site_pk=None, *args, **kwargs):
        """
        Delete multiple objects by providing a list of IDs in the request body.

        Accepts a JSON list of integer IDs, e.g.::

            DELETE /api/sites/1/devices/
            [1, 2, 3]

        Each object's permissions are checked before deletion.
        """
        ids = request.data

        if not isinstance(ids, list):
            raise exc.BadRequest(
                "Bulk delete expects a list of IDs in the request body."
            )

        if len(ids) == 0:
            raise exc.BadRequest("No IDs provided for bulk delete.")

        # Validate all IDs are integers (reject booleans, which are int subclass)
        for id_ in ids:
            if isinstance(id_, bool) or not isinstance(id_, int):
                raise exc.BadRequest(
                    "All IDs must be integers. Got: %r" % (id_,)
                )

        # Deduplicate while preserving order to prevent multiple Change
        # records for the same object.
        ids = list(dict.fromkeys(ids))

        queryset = self.get_queryset()
        if site_pk is not None:
            queryset = queryset.filter(site=site_pk)

        # Look up all objects and check permissions
        objects = []
        for id_ in ids:
            try:
                obj = queryset.get(pk=id_)
            except exc.ObjectDoesNotExist:
                self.not_found(id_, site_pk)

            self.check_object_permissions(request, obj)
            objects.append(obj)

        # Perform deletion atomically so partial failures don't leave
        # the database in a half-deleted state.
        with transaction.atomic():
            for obj in objects:
                self.perform_destroy(obj)

        return Response(status=status_codes.HTTP_204_NO_CONTENT)


class SiteViewSet(NsotViewSet):
    """
    API endpoint that allows Sites to be viewed or edited.
    """

    queryset = models.Site.objects.all()
    serializer_class = serializers.SiteSerializer
    filterset_fields = ("name",)


class ValueViewSet(NsotViewSet):
    """
    API endpoint that allows Attribute Values to be viewed or edited.
    """

    queryset = models.Value.objects.all()
    serializer_class = serializers.ValueSerializer
    filterset_fields = ("name", "value", "resource_name", "resource_id")

    def get_serializer_class(self):
        if self.request.method == "POST":
            return serializers.ValueCreateSerializer
        return self.serializer_class


class NsotBulkUpdateModelMixin(bulk_mixins.BulkUpdateModelMixin):
    """
    The default mixin isn't using super() so multiple-inheritance breaks. This
    fixes it for our use-case.
    """

    def perform_update(self, serializer):
        super(bulk_mixins.BulkUpdateModelMixin, self).perform_update(
            serializer
        )


class ResourceViewSet(
    NsotBulkUpdateModelMixin, NsotViewSet, bulk_mixins.BulkCreateModelMixin
):
    """
    Resource views that include set query list endpoints.
    """

    @action(methods=["get"], detail=False)
    def query(self, request, site_pk=None, *args, **kwargs):
        """Perform a set query."""
        query = request.query_params.get("query", "")
        unique = qpbool(request.query_params.get("unique", False))
        qs = self.queryset.set_query(query, site_id=site_pk, unique=unique)
        objects = self.filter_queryset(qs)
        return self.list(request, queryset=objects, *args, **kwargs)

    def get_resource_object(self, pk, site_pk):
        """Return a resource object based on pk or site_pk."""
        # FIXME(jathan): Revisit this after we've seen if we can need to get
        # rid of the overloaded .get_queryset() methods on the ResourceViewSet
        # subclasses, or provide them with some sort of "don't filter queries"
        # flag. TBD.

        # Backup the original kwargs
        orig_kwargs = self.kwargs.copy()
        self.kwargs["pk"] = pk
        self.kwargs["site_pk"] = site_pk

        # Get our object and restore the original kwargs
        obj = self.get_object()
        self.kwargs = orig_kwargs

        return obj


class AttributeViewSet(ResourceViewSet):
    """
    API endpoint that allows Attributes to be viewed or edited.
    """

    queryset = models.Attribute.objects.all()
    serializer_class = serializers.AttributeSerializer
    filterset_class = filters.AttributeFilter

    def get_serializer_class(self):
        if self.request.method == "POST":
            return serializers.AttributeCreateSerializer
        if self.request.method in ("PUT", "PATCH"):
            return serializers.AttributeUpdateSerializer
        return self.serializer_class


class DeviceViewSet(ResourceViewSet):
    """
    API endpoint that allows Devices to be viewed or edited.
    """

    queryset = models.Device.objects.all()
    serializer_class = serializers.DeviceSerializer
    filterset_class = filters.DeviceFilter
    lookup_value_regex = settings.DEVICE_NAME_PATTERN
    natural_key = "hostname"

    def get_serializer_class(self):
        if self.request.method == "POST":
            return serializers.DeviceCreateSerializer
        if self.request.method == "PUT":
            return serializers.DeviceUpdateSerializer
        if self.request.method == "PATCH":
            return serializers.DevicePartialUpdateSerializer

        return self.serializer_class

    @action(methods=["get"], detail=True)
    def interfaces(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return all interfaces for this Device."""
        device = self.get_resource_object(pk, site_pk)
        interfaces = device.interfaces.all()

        return self.list(
            request,
            queryset=interfaces,
            serializer_class=serializers.InterfaceSerializer,
            *args,
            **kwargs,
        )

    @action(methods=["get"], detail=True)
    def circuits(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return a list of Circuits for this Device"""
        device = self.get_resource_object(pk, site_pk)
        circuits = device.circuits

        return self.list(
            request,
            queryset=circuits,
            serializer_class=serializers.CircuitSerializer,
            *args,
            **kwargs,
        )


class NetworkViewSet(ResourceViewSet):
    """
    API endpoint that allows Networks to be viewed or edited.
    """

    queryset = models.Network.objects.all()
    serializer_class = serializers.NetworkSerializer
    filterset_class = filters.NetworkFilter
    lookup_value_regex = "[a-fA-F0-9:./]+"
    natural_key = "cidr"

    # Thd default number of networks that is returned
    DEFAULT_NETWORK_NUM = 1

    def allocate_networks(self, networks, site_pk, state="allocated"):
        site = models.Site.objects.get(pk=site_pk)
        for n in networks:
            obj = models.Network(cidr=n, site=site, state=state)
            obj.save()
            models.Change.objects.create(
                obj=obj, user=self.request.user, event="Create"
            )

    def get_serializer_class(self):
        if self.request.method == "POST":
            return serializers.NetworkCreateSerializer
        if self.request.method == "PUT":
            return serializers.NetworkUpdateSerializer
        if self.request.method == "PATCH":
            return serializers.NetworkPartialUpdateSerializer

        return self.serializer_class

    def get_natural_key_kwargs(self, filter_value):
        """Return a dict of kwargs for natural_key lookup."""
        return cidr_to_dict(filter_value)

    @action(methods=["get"], detail=False)
    def query(self, request, site_pk=None, *args, **kwargs):
        """Override base query to inherit filtering by query params."""
        self.queryset = self.get_queryset()
        return super().query(request, site_pk, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def closest_parent(self, request, pk=None, site_pk=None, *args, **kwargs):
        """
        Return the closest matching parent of this Network even if it doesn't
        exist in the database.
        """
        prefix_length = request.query_params.get("prefix_length", 0)

        # Get closest parent or 404
        try:
            network = models.Network.objects.get_closest_parent(
                pk, prefix_length=prefix_length, site=site_pk
            )
        except models.Network.DoesNotExist:
            self.not_found(pk, site_pk)
        else:
            pk = network.id  # Use the id of the network we found

        return self.retrieve(request, pk, site_pk, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def subnets(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return subnets of this Network."""
        network = self.get_resource_object(pk, site_pk)

        params = request.query_params
        include_networks = qpbool(params.get("include_networks", True))
        include_ips = qpbool(params.get("include_ips", True))
        direct = qpbool(params.get("direct", False))

        networks = network.subnets(
            include_networks=include_networks,
            include_ips=include_ips,
            direct=direct,
        )

        return self.list(request, queryset=networks, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def supernets(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return supernets of this Network."""
        network = self.get_resource_object(pk, site_pk)

        params = request.query_params
        direct = qpbool(params.get("direct", False))

        networks = network.supernets(direct=direct)

        return self.list(request, queryset=networks, *args, **kwargs)

    @action(methods=["get", "post"], detail=True)
    def next_network(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return next available networks from this Network."""
        network = self.get_resource_object(pk, site_pk)
        params = request.query_params

        prefix_length = params.get("prefix_length")
        num = params.get("num", self.DEFAULT_NETWORK_NUM)
        strict = qpbool(params.get("strict_allocation", False))

        networks = network.get_next_network(
            prefix_length, num, strict, as_objects=False
        )

        if request.method == "POST":
            if qpbool(params.get("reserve", False)):
                state = models.Network.RESERVED
            else:
                state = models.Network.ALLOCATED
            self.allocate_networks(networks, site_pk, state)
        return self.success(networks)

    @action(methods=["get", "post"], detail=True)
    def next_address(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return next available IPs from this Network."""
        network = self.get_resource_object(pk, site_pk)
        params = request.query_params

        num = params.get("num", self.DEFAULT_NETWORK_NUM)
        strict = qpbool(params.get("strict_allocation", False))
        addresses = network.get_next_address(num, strict, as_objects=False)

        if request.method == "POST":
            if qpbool(params.get("reserve", False)):
                state = models.Network.RESERVED
            else:
                state = models.Network.ALLOCATED
            self.allocate_networks(addresses, site_pk, state)
        return self.success(addresses)

    @action(methods=["get"], detail=True)
    def ancestors(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return ancestors of this Network."""
        network = self.get_resource_object(pk, site_pk)
        ascending = qpbool(request.query_params.get("ascending", False))
        ancestors = network.get_ancestors(ascending=ascending)

        return self.list(request, queryset=ancestors, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def children(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return the immediate children of this Network."""
        network = self.get_resource_object(pk, site_pk)
        children = network.get_children()

        return self.list(request, queryset=children, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def descendants(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return descendants of this Network."""
        network = self.get_resource_object(pk, site_pk)
        descendants = network.get_descendants()

        return self.list(request, queryset=descendants, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def parent(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return the parent of this Network."""
        network = self.get_resource_object(pk, site_pk)
        parent = network.parent
        pk = network.parent_id if parent is not None else None

        return self.retrieve(request, pk, site_pk, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def root(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return the parent of all ancestors for this Network."""
        network = self.get_resource_object(pk, site_pk)
        root = network.get_root()
        pk = root.id if root is not None else None

        return self.retrieve(request, pk, site_pk, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def siblings(self, request, pk=None, site_pk=None, *args, **kwargs):
        """
        Return Networks with the same parent. Root nodes are
        siblings to other root nodes.
        """
        network = self.get_resource_object(pk, site_pk)
        include_self = qpbool(request.query_params.get("include_self", False))
        descendents = network.get_siblings(include_self=include_self)

        return self.list(request, queryset=descendents, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def assignments(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return the interface assignments for this Network."""
        network = self.get_resource_object(pk, site_pk)
        assignments = network.assignments.all()

        return self.list(
            request,
            queryset=assignments,
            serializer_class=serializers.AssignmentSerializer,
            *args,
            **kwargs,
        )

    @action(methods=["get"], detail=False)
    def reserved(self, request, site_pk=None, *args, **kwargs):
        """Display all reserved Networks."""
        objects = models.Network.objects.reserved()
        return self.list(request, queryset=objects, *args, **kwargs)

    # Shoutout to jathanism for suggesting the move of perform_destroy(()
    # to the NetworkViewSet to make it more explicit.
    def perform_destroy(self, instance):
        """
        Overload default to handle non-serializer exceptions, and log Change
        events.
        :param instance:
            Model instance to delete
        """
        log.debug("NetworkViewSet.perform_destroy() obj = %r", instance)
        change = models.Change.objects.create(
            obj=instance, user=self.request.user, event="Delete"
        )
        force_delete = qpbool(
            self.request.query_params.get("force_delete", False)
        )
        try:
            instance.delete(force_delete=force_delete)
        except exc.ProtectedError as err:
            change.delete()
            raise exc.Conflict(err.args[0])


class InterfaceViewSet(ResourceViewSet):
    """
    API endpoint that allows Interfaces to be viewed or edited.
    """

    queryset = models.Interface.objects.all()
    serializer_class = serializers.InterfaceSerializer
    filterset_class = filters.InterfaceFilter
    # Match on device_hostname:name or pk id
    # Being pretty vague here, so as to be minimally prescriptive
    lookup_value_regex = "[a-zA-Z0-9:./-]*[0-9]"
    natural_key = "name_slug"

    def get_serializer_class(self):
        if self.request.method == "POST":
            return serializers.InterfaceCreateSerializer
        if self.request.method == "PUT":
            return serializers.InterfaceUpdateSerializer
        if self.request.method == "PATCH":
            return serializers.InterfacePartialUpdateSerializer
        return self.serializer_class

    @action(methods=["get"], detail=True)
    def addresses(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return a list of addresses for this Interface."""
        interface = self.get_resource_object(pk, site_pk)
        addresses = interface.addresses.all()
        return self.list(
            request,
            queryset=addresses,
            serializer_class=serializers.NetworkSerializer,
            *args,
            **kwargs,
        )

    @action(methods=["get"], detail=True)
    def assignments(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return a list of information about my assigned addresses."""
        interface = self.get_resource_object(pk, site_pk)
        assignments = interface.assignments.all()
        return self.list(
            request,
            queryset=assignments,
            serializer_class=serializers.AssignmentSerializer,
            *args,
            **kwargs,
        )

    @action(methods=["get"], detail=True)
    def networks(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return all the containing Networks for my assigned addresses."""
        interface = self.get_resource_object(pk, site_pk)
        return self.list(
            request,
            queryset=interface.networks,
            serializer_class=serializers.NetworkSerializer,
            *args,
            **kwargs,
        )

    @action(methods=["get"], detail=True)
    def parent(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return the parent of this Interface."""
        interface = self.get_resource_object(pk, site_pk)
        parent = interface.parent
        pk = interface.parent_id if parent is not None else None
        return self.retrieve(request, pk, site_pk, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def ancestors(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return all the ancestors of this Interface."""
        interface = self.get_resource_object(pk, site_pk)
        return self.list(
            request, queryset=interface.get_ancestors(), *args, **kwargs
        )

    @action(methods=["get"], detail=True)
    def children(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return all the immediate children of this Interface."""
        interface = self.get_resource_object(pk, site_pk)
        return self.list(
            request, queryset=interface.get_children(), *args, **kwargs
        )

    @action(methods=["get"], detail=True)
    def descendants(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return all the descendants of this Interface."""
        interface = self.get_resource_object(pk, site_pk)
        return self.list(
            request, queryset=interface.get_descendants(), *args, **kwargs
        )

    @action(methods=["get"], detail=True)
    def siblings(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return all the siblings of this Interface."""
        interface = self.get_resource_object(pk, site_pk)
        return self.list(
            request, queryset=interface.get_siblings(), *args, **kwargs
        )

    @action(methods=["get"], detail=True)
    def root(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return the root of the tree this Interface is part of."""
        interface = self.get_resource_object(pk, site_pk)
        root = interface.get_root()
        pk = root.id
        return self.retrieve(request, pk, site_pk, *args, **kwargs)

    @action(methods=["get"], detail=True)
    def circuit(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return the Circuit I am associated with"""
        interface = self.get_resource_object(pk, site_pk)
        try:
            cir = serializers.CircuitSerializer(interface.circuit)
            return self.success(cir.data)
        except models.Circuit.DoesNotExist:
            msg = "No Circuit found at Interface (site_id, id) = ({}, {})"
            msg = msg.format(site_pk, pk)
            self.not_found(pk, msg=msg)


class CircuitViewSet(ResourceViewSet):
    """
    API endpoint that allows Circuits to be viewed or edited.
    """

    queryset = models.Circuit.objects.all()
    serializer_class = serializers.CircuitSerializer
    filterset_class = filters.CircuitFilter
    natural_key = "name_slug"

    # TODO(jathan): Revisit this if and when we upgrade or replace
    # django-rest-framework-bulk==0.2.1
    def bulk_update(self, request, *args, **kwargs):
        """
        Workaround for bulk update of objects with unique constraint.

        At this time this is only required by the Circuit object, which is why
        it is only defined here.

        Credit: https://github.com/miki725/django-rest-framework-bulk/issues/30
        Source: http://bit.ly/2HcyNnG
        """
        partial = kwargs.pop("partial", False)

        # restrict the update to the filtered queryset
        serializer = self.get_serializer(
            self.filter_queryset(self.get_queryset()),
            data=request.data,
            many=True,
            partial=partial,
        )

        validated_data = []
        validation_errors = []

        for item in request.data:
            item_serializer = self.get_serializer(
                get_object_or_404(
                    self.filter_queryset(self.get_queryset()), pk=item["id"]
                ),
                data=item,
                partial=partial,
            )
            if not item_serializer.is_valid():
                validation_errors.append(item_serializer.errors)

            obj_data = item_serializer.validated_data
            # By default validated_data does not have `id`, so adding it in
            # validated_data of each item
            obj_data["id"] = item["id"]
            validated_data.append(obj_data)

        if validation_errors:
            raise exc.ValidationError(validation_errors)

        serializer._validated_data = validated_data
        self.perform_bulk_update(serializer)

        return self.success(serializer.data)

    def get_serializer_class(self):
        if self.request.method == "POST":
            return serializers.CircuitCreateSerializer
        if self.request.method == "PUT":
            return serializers.CircuitUpdateSerializer
        if self.request.method == "PATCH":
            return serializers.CircuitPartialUpdateSerializer

        return self.serializer_class

    @action(methods=["get"], detail=True)
    def addresses(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return a list of addresses for the interfaces on this Circuit."""
        circuit = self.get_resource_object(pk, site_pk)
        addresses = circuit.addresses

        return self.list(
            request,
            queryset=addresses,
            serializer_class=serializers.NetworkSerializer,
            *args,
            **kwargs,
        )

    @action(methods=["get"], detail=True)
    def devices(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return a list of devices for this Circuit."""
        circuit = self.get_resource_object(pk, site_pk)
        devices = circuit.devices

        return self.list(
            request,
            queryset=devices,
            serializer_class=serializers.DeviceSerializer,
            *args,
            **kwargs,
        )

    @action(methods=["get"], detail=True)
    def interfaces(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return a list of interfaces for this Circuit."""
        circuit = self.get_resource_object(pk, site_pk)
        interfaces = circuit.interfaces

        return self.list(
            request,
            queryset=interfaces,
            serializer_class=serializers.InterfaceSerializer,
            *args,
            **kwargs,
        )


class ProtocolTypeViewSet(NsotViewSet):
    """
    API endpoint that allows ProtocolTypes to be viewed or edited.
    """

    queryset = models.ProtocolType.objects.all()
    serializer_class = serializers.ProtocolTypeSerializer
    filterset_class = filters.ProtocolTypeFilter
    natural_key = "name"


class ProtocolViewSet(ResourceViewSet):
    """
    API endpoint that allows Protocols to be viewed or edited.
    """

    queryset = models.Protocol.objects.all()
    serializer_class = serializers.ProtocolSerializer
    filterset_class = filters.ProtocolFilter

    def get_serializer_class(self):
        if self.request.method == "POST":
            return serializers.ProtocolCreateSerializer
        if self.request.method == "PUT":
            return serializers.ProtocolUpdateSerializer
        if self.request.method == "PATCH":
            return serializers.ProtocolPartialUpdateSerializer

        return self.serializer_class


#: Namedtuple for retrieving pk and user object of current user.
UserPkInfo = namedtuple("UserPkInfo", "user pk")


class UserViewSet(BaseNsotViewSet, mixins.CreateModelMixin):
    """
    This viewset automatically provides `list` and `detail` actins.
    """

    queryset = get_user_model().objects.all()
    serializer_class = serializers.UserSerializer
    filterset_fields = ("email",)

    def get_user_and_pk(self, request, pk=None, site_pk=None):
        # If pk is 0, return the current user.
        if int(pk) == 0:
            pk = request.user.pk  # Authenticated user

        # Try to get the requested user
        try:
            user = self.queryset.get(pk=pk)
        except exc.ObjectDoesNotExist:
            self.not_found(pk, site_pk)

        return UserPkInfo(user, pk)

    def retrieve(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Retreive a single user."""
        user, pk = self.get_user_and_pk(request, pk, site_pk)

        params = request.query_params
        with_secret_key = params.get("with_secret_key", None)

        # If with_secret_key is set, confirm that the requested user object
        # matches the current user.
        if with_secret_key is not None:
            if user != request.user:
                raise exc.Forbidden(
                    "Can't access secret_key of user that isn't you."
                )
            kwargs["with_secret_key"] = qpbool(with_secret_key)

        return super().retrieve(request, pk, site_pk, *args, **kwargs)

    @action(methods=["post"], detail=True)
    def rotate_secret_key(self, request, pk=None, *args, **kwargs):
        user, pk = self.get_user_and_pk(request, pk)

        if user != request.user:
            raise exc.Forbidden(
                "Can't access secret_key of user that isn't you."
            )

        user.rotate_secret_key()
        return self.success(user.secret_key)


class NotFoundViewSet(viewsets.GenericViewSet):
    """Catchall for bad API endpoints."""

    permission_classes = (permissions.IsAuthenticated,)
    schema = None

    def get_queryset(self):
        return None

    def get(self, *args, **kwargs):
        raise exc.NotFound

    def list(self, *args, **kwargs):
        raise exc.NotFound

    def get_serializer_class(self):
        return None


class AuthTokenLoginView(APIView):
    permission_classes = ()

    def post(self, request, *args, **kwargs):
        serializer = serializers.AuthTokenSerializer(data=request.data)

        if serializer.is_valid():
            user = serializer.validated_data["user"]
            data = {"auth_token": user.generate_auth_token()}

            if request.version == settings.NSOT_API_VERSION:
                return Response(data)

            return Response(
                OrderedDict(
                    [
                        ("status", "ok"),
                        ("data", data),
                    ]
                )
            )
        raise exc.Unauthorized(serializer.errors)


class AuthTokenVerifyView(APIView):
    authentication_classes = (auth.AuthTokenAuthentication,)
    permission_classes = ()

    def post(self, request, *args, **kwargs):
        if request.version == settings.NSOT_API_VERSION:
            return Response(True)

        return Response(
            OrderedDict(
                [
                    ("status", "ok"),
                    ("data", True),
                ]
            )
        )



"""
Utilities used across the project.
"""

# Core
# Stats
from . import core, stats
from .core import *  # noqa
from .stats import *  # noqa

__all__ = []
__all__.extend(core.__all__)
__all__.extend(stats.__all__)



"""
Command for starting up an authenticating reverse proxy for use in development.

Please, don't use me in production!
"""

import getpass
import http.server

from django.conf import settings

from nsot.util.commands import CommandError, NsotCommand


class Command(NsotCommand):
    help = "Start an authenticating reverse proxy for use in development."

    def add_arguments(self, parser):
        parser.add_argument(
            "username",
            nargs="?",
            default=getpass.getuser(),
            help="Username used for authentication.",
        )
        parser.add_argument(
            "-a",
            "--address",
            type=str,
            default=settings.NSOT_HOST,
            help="Address to listen on.",
        )
        parser.add_argument(
            "-d",
            "--domain",
            type=str,
            default="localhost",
            help="Domain for user account.",
        )
        parser.add_argument(
            "-H",
            "--auth-header",
            type=str,
            default=settings.USER_AUTH_HEADER,
            help="HTTP user auth header name.",
        )
        parser.add_argument(
            "-P",
            "--backend-port",
            type=int,
            default=settings.NSOT_PORT,
            help="Port to proxy to.",
        )
        parser.add_argument(
            "-p",
            "--listen-port",
            type=int,
            default=settings.NSOT_PORT + 1,
            help="Port to listen on.",
        )

    def handle(self, **options):
        username = options.get("username")

        try:
            from mrproxy import UserProxyHandler
        except ImportError:
            raise SystemExit(
                "mrproxy is required for the user proxy. Please see "
                "README.rst."
            )

        class ServerArgs:
            """Argument container for http service."""

            def __init__(self, backend_port, username, auth_header):
                self.backend_port = backend_port
                self.header = ["%s: %s" % (auth_header, username)]

        username = "%s@%s" % (username, options.get("domain"))
        address = options.get("address")
        auth_header = options.get("auth_header")
        backend_port = options.get("backend_port")
        listen_port = options.get("listen_port")

        # Try to start the server
        try:
            server = http.server.HTTPServer(
                (address, listen_port), UserProxyHandler
            )
        except OSError as err:
            raise CommandError(err)
        else:
            server.args = ServerArgs(backend_port, username, auth_header)

        # Run until we hit ctrl-C
        try:
            print(
                "Starting proxy on %s %s => %s, auth '%s: %s'"
                % (address, backend_port, listen_port, auth_header, username)
            )
            server.serve_forever()
        except KeyboardInterrupt:
            print("Bye!")



import logging

from django.conf import settings
from django.core.cache import cache as djcache
from django.db import models
from django.utils import timezone

from .. import exc, fields, util, validators
from . import constants
from .assignment import Assignment
from .circuit import Circuit
from .device import Device
from .network import Network
from .resource import Resource

log = logging.getLogger(__name__)


class Interface(Resource):
    """A network interface."""

    # Derive site from device
    site_source_field = "device"

    # Don't validate unique in save() - we want IntegrityError for 409 CONFLICT
    _validate_unique_on_save = False

    # if_name
    # SNMP: ifName
    # if_description
    # SNMP: ifDescr
    name = models.CharField(
        max_length=255,
        null=False,
        db_index=True,
        help_text="The name of the interface as it appears on the Device.",
    )

    # This doesn't use the built-in SlugField since we're doing our own
    # slugification (django.utils.text.slugify() is too agressive)
    name_slug = models.CharField(
        db_index=True,
        editable=False,
        max_length=255,
        null=True,
        unique=True,
        help_text=(
            "Slugified version of the name field, used for the natural key"
        ),
    )

    # if_addr
    # m2m Network object /32 or /128
    addresses = models.ManyToManyField(
        "Network",
        db_index=True,
        related_name="addresses",
        through="Assignment",
        help_text="Network addresses assigned to this Interface",
    )

    # if_alias - String of interface description
    # SNMP: ifAlias
    description = models.CharField(
        max_length=255,
        default="",
        blank=True,
        null=False,
        help_text="A brief yet helpful description.",
    )

    # server_id
    device = models.ForeignKey(
        "Device",
        db_index=True,
        related_name="interfaces",
        null=False,
        on_delete=models.CASCADE,
        verbose_name="Device",
        help_text="Unique ID of the connected Device.",
    )

    # Cached hostname of the associated device
    device_hostname = models.CharField(
        max_length=255,
        null=False,
        blank=True,
        db_index=True,
        editable=False,
        help_text=(
            "The hostname of the Device to which the interface is bound. "
            "(Internal use only)"
        ),
    )

    # if_type - Integer of interface type id (Ethernet, LAG, etc.)
    # SNMP: ifType
    type = models.IntegerField(
        "Interface Type",
        choices=settings.INTERFACE_TYPE_CHOICES,
        default=settings.INTERFACE_DEFAULT_TYPE,
        null=False,
        db_index=True,
        help_text="If not provided, defaults to 'ethernet'.",
    )

    # if_physical_address - Integer of hex MAC address
    # SNMP: ifPhysAddress
    mac_address = fields.MACAddressField(
        "MAC Address",
        blank=True,
        db_index=True,
        null=True,
        default=int(settings.INTERFACE_DEFAULT_MAC),
        help_text=(
            "If not provided, defaults to %s." % settings.INTERFACE_DEFAULT_MAC
        ),
    )

    # if_speed - Should not be used. Caps at 4.3GB (2^32)
    # SNMP: ifSpeed

    # if_high_speed - Integer of Mbps of interface (e.g. 20000 for 20 Gbps)
    # SNMP: ifHighSpeed
    speed = models.IntegerField(
        blank=True,
        db_index=True,
        default=settings.INTERFACE_DEFAULT_SPEED,
        null=True,
        help_text=(
            "Integer of Mbps of interface (e.g. 20000 for 20 Gbps). If not "
            "provided, defaults to %s." % settings.INTERFACE_DEFAULT_SPEED
        ),
    )

    # Maximum Transmission Unit (MTU) in bytes
    mtu = models.IntegerField(
        blank=True,
        db_index=True,
        default=None,
        null=True,
        help_text="MTU (Maximum Transmission Unit) in bytes.",
    )

    parent = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="children",
        default=None,
        db_index=True,
        on_delete=models.PROTECT,
        verbose_name="Parent",
        help_text="Unique ID of the parent Interface.",
    )

    # We are currently inferring the site_id from the parent Device in the
    # .save() method. We don't want to even care about the site_id , but it
    # simplifies managing them this way.
    site = models.ForeignKey(
        "Site",
        db_index=True,
        related_name="interfaces",
        on_delete=models.PROTECT,
        help_text="Unique ID of the Site this Interface is under.",
    )

    # Where list of assigned addresses is cached.
    _addresses_cache = fields.JSONField(null=False, blank=True, default=[])

    # Where list of attached networks is cached.
    _networks_cache = fields.JSONField(null=False, blank=True, default=[])

    def __init__(self, *args, **kwargs):
        self._set_addresses = kwargs.pop("addresses", None)
        super().__init__(*args, **kwargs)

    ##########################################
    # THESE WILL BE IMPLEMENTED AS ATTRIBUTES
    ##########################################

    # Using the "indexing strategy" concept, which I think will be pluggable.
    # if_index - Integer of SNMP interface index on device
    # SNMP: ifIndex
    # snmp_index

    # if_parent_index
    # snmp_parent_index

    # LLDP adjacencies
    # These should be deployed as attributes.
    # lldp_remote_port_desc
    # lldp_remote_system_name

    def __str__(self):
        return self.name_slug

    class Meta:
        unique_together = ("device", "name")
        indexes = [
            models.Index(fields=["device_hostname", "name"]),
        ]

    @property
    def networks(self):
        """Return all the parent Networks for my addresses."""
        return Network.objects.filter(
            id__in=list(self.addresses.values_list("parent", flat=True))
        ).distinct()

    @property
    def circuit(self):
        """Return the Circuit I am associated with"""
        try:
            return self.circuit_a
        except Circuit.DoesNotExist:
            try:
                return self.circuit_z
            except Circuit.DoesNotExist:
                raise

    def _purge_addresses(self):
        """Delete all of my addresses (and therefore assignments)."""
        self.addresses.all().delete()
        self.clean_addresses()  # Always re-cache after we purge addresses..

    def _purge_assignments(self):
        """Delete all of my assignments, leaving the Network objects intact."""
        self.assignments.all().delete()
        self.clean_addresses()  # Always re-cache after we purge assignments.

    def assign_address(self, cidr):
        """
        Assign an address to this interface.

        Must have prefix of /32 (IPv4) or /128 (IPv6).

        :param cidr:
            IPv4/v6 CIDR host address or Network object
        """
        cidr = validators.validate_host_address(cidr)
        try:
            address = Network.objects.get_by_address(cidr, site=self.site)
        except Network.DoesNotExist:
            address = Network.objects.create(cidr=cidr, site=self.site)
            created = True
        else:
            created = False

        # If we've created a Network, and assignment fails, then we need to
        # make sure that we delete that Network.
        try:
            return Assignment.objects.create(interface=self, address=address)
        except exc.ValidationError:
            # Clean up the address if we created one..
            if created:
                address.delete()
            raise

    def set_addresses(self, addresses, overwrite=False, partial=False):
        """
        Explicitly assign a list of addresses to this Interface.

        :param addresses:
            A list of CIDRs

        :param overwrite:
            Whether to purge existing assignments before assigning.

        :param partial:
            Whether this is a partial update.
        """
        log.debug("Interface.set_addresses() addresses = %r", addresses)

        # If no addresses and it's a partial update, NOOP.
        if addresses is None and partial:
            return

        if not isinstance(addresses, list):
            raise exc.ValidationError(
                {"addresses": f"Expected list but received {type(addresses)}"}
            )

        if overwrite:
            self._purge_assignments()

        # Keep track of addresses that are already assigned so we don't try to
        # assign them again (which would result in an error).
        existing_addresses = self.get_addresses()

        inserts = []
        for cidr in addresses:
            # Don't assign an address that already exists.
            if cidr in existing_addresses:
                continue

            address = validators.validate_cidr(cidr)
            inserts.append(str(address))

        for insert in inserts:
            self.assign_address(insert)

        self.clean_addresses()

    def get_ancestors(self):
        """Return all ancestors of an Interface."""
        p = self.parent
        ancestors = []
        while p is not None:
            ancestors.append(p.id)
            p = p.parent
        return Interface.objects.filter(id__in=ancestors)

    def get_children(self):
        """Return the immediate children of an Interface."""
        return Interface.objects.filter(parent=self)

    def get_descendants(self):
        """Return all the descendants of an Interface."""
        s = list(self.get_children())
        descendants = []
        while len(s) > 0:
            top = s.pop()
            descendants.append(top.id)
            for c in top.get_children():
                s.append(c)
        return Interface.objects.filter(id__in=descendants)

    def get_root(self):
        """Return the parent of all ancestors of an Interface."""
        root = self
        while root.parent is not None:
            root = root.parent
        return root

    def get_siblings(self):
        """
        Return Interfaces with the same parent and device id as an Interface.
        """
        return Interface.objects.filter(
            parent=self.parent, device=self.device
        ).exclude(id=self.id)

    def get_assignments(self):
        """Return a list of information about my assigned addresses."""
        return [a.to_dict() for a in self.assignments.all()]

    def get_addresses(self):
        """Return a list of assigned addresses."""
        return self._addresses_cache

    def get_networks(self):
        """Return a list of attached Networks."""
        return self._networks_cache

    def get_mac_address(self):
        """Return a serializable representation of mac_address."""
        if self.mac_address is None:
            return None
        return str(self.mac_address)

    def clean_addresses(self):
        """Make sure that addresses/networks are saved as JSON."""
        addresses = [a.cidr for a in self.addresses.iterator()]
        self._addresses_cache = addresses

        networks = [n.cidr for n in self.networks.iterator()]
        self._networks_cache = networks

    def clean_name(self, value):
        """Enforce name."""
        return validators.validate_name(value)

    def clean_speed(self, value):
        """Enforce valid speed."""
        if value is None:
            return value

        # We don't want floats because they can be misleading, also Django's
        # IntegerField will cast a float to an int, which loses precision.
        # TODO (jathan): Reconsider this as a float? Maybe? We might not care
        # about things like 1.544 (T1) anymore...
        if isinstance(value, float):
            raise exc.ValidationError({"speed": "Speed must be an integer."})

        try:
            value = int(value)
        except ValueError:
            raise exc.ValidationError({"speed": "Invalid speed: %r" % value})
        else:
            return value

    def clean_mtu(self, value):
        """Enforce valid MTU."""
        if value is None:
            return value

        if isinstance(value, float):
            raise exc.ValidationError({"mtu": "MTU must be an integer."})

        try:
            value = int(value)
        except (ValueError, TypeError):
            raise exc.ValidationError({"mtu": "Invalid MTU: %r" % value})

        if value < 68 or value > 65535:
            raise exc.ValidationError(
                {"mtu": "MTU must be between 68 and 65535."}
            )

        return value

    def clean_type(self, value):
        """Enforce valid type. Accepts integer IDs or string names."""
        if value is None:
            value = settings.INTERFACE_DEFAULT_TYPE
        if isinstance(value, str):
            value = constants.INTERFACE_TYPE_BY_NAME.get(value, value)
        if value not in constants.INTERFACE_TYPES:
            raise exc.ValidationError({"type": "Invalid type: %r" % value})

        return value

    def clean_mac_address(self, value):
        """Enforce valid mac_address."""
        return validators.validate_mac_address(value)

    def clean_device_hostname(self, device):
        """Extract hostname from device"""
        return device.hostname

    def clean_name_slug(self, value=None):
        """Slugify the interface name into natural key."""
        if value is None:
            value = util.slugify_interface(
                device_hostname=self.device_hostname, name=self.name
            )
        return value

    def clean_parent(self, parent):
        if parent is None:
            return parent
        if parent.device_hostname != self.device_hostname:
            raise exc.ValidationError(
                {
                    "parent": (
                        "Parent's device does not match device with host "
                        "name %r" % self.device_hostname
                    )
                }
            )
        return parent

    def clean_fields(self, exclude=None):
        self.site_id = self.clean_site(self.site_id)
        self.name = self.clean_name(self.name)
        self.type = self.clean_type(self.type)
        self.speed = self.clean_speed(self.speed)
        self.mtu = self.clean_mtu(self.mtu)
        self.mac_address = self.clean_mac_address(self.mac_address)
        self.device_hostname = self.clean_device_hostname(self.device)
        self.parent = self.clean_parent(self.parent)
        self.name_slug = self.clean_name_slug()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

        # This is so that we can set the addresses on create/update, but if
        # the object is new, make sure that it doesn't persist if addresses
        # fail.
        if self._set_addresses is not None:
            try:
                # And set the attributes (if any)
                self.set_addresses(self._set_addresses)
            except exc.ValidationError:
                # If addresses fail validation, and I'm a new object, delete
                # myself and re-raise the error.
                if self._is_new:  # This is set by Resource.save()
                    self.delete()
                raise

    def to_dict(self):
        return {
            "id": self.id,
            "parent_id": self.parent_id,
            "parent": self.parent and self.parent.name_slug,
            "name": self.name,
            "name_slug": self.name_slug,
            "device": self.device_id,
            "device_hostname": self.device_hostname,
            "description": self.description,
            "addresses": self.get_addresses(),
            "networks": self.get_networks(),
            "mac_address": self.get_mac_address(),
            "speed": self.speed,
            "mtu": self.mtu,
            "type": self.type,
            "type_name": self.get_type_display(),
            "attributes": self.get_attributes(),
        }


# Signals
def change_api_updated_at(sender=None, instance=None, *args, **kwargs):
    """Anytime the API is updated, invalidate the cache."""
    djcache.set("api_updated_at_timestamp", timezone.now())


def update_device_interfaces(sender, instance, **kwargs):
    """Anytime a device is saved, update device_hostname on its interfaces"""
    interfaces = Interface.objects.filter(device=instance)
    for interface in interfaces.iterator():
        interface.device_hostname = instance.hostname
        interface.name_slug = None  # null so model validation resets name_slug
        interface.save()


models.signals.post_save.connect(
    change_api_updated_at,
    sender=Interface,
    dispatch_uid="invalidate_cache_post_save_interface",
)
models.signals.post_delete.connect(
    change_api_updated_at,
    sender=Interface,
    dispatch_uid="invalidate_cache_post_delete_interface",
)
models.signals.post_save.connect(
    update_device_interfaces,
    sender=Device,
    dispatch_uid="update_interface_post_save_device",
)



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django_extensions.db.fields.json

def migrate_device_fields(apps, schema_editor):
    """
    Migrate new Device fields.
    """
    Device = apps.get_model("nsot", "Device")
    Device_temp = apps.get_model("nsot", "Device_temp")
    for dev in Device.objects.iterator():
        dev_tmp = Device_temp.objects.create(
            id=dev.resource_ptr_id,
            hostname=dev.hostname,
            _attributes_cache=dev._attributes,
            site=dev.site,
        )

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0015_move_attribute_fields"),
    ]

    operations = [
        # Device _attributes_cache, new_id
        migrations.RunPython(migrate_device_fields),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations, models

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0023_auto_20151008_1351"),
    ]

    operations = [
        migrations.AddField(
            model_name="network",
            name="state",
            field=models.CharField(
                default="allocated",
                max_length=20,
                db_index=True,
                choices=[
                    ("allocated", "Allocated"),
                    ("assigned", "Assigned"),
                    ("orphaned", "Orphaned"),
                    ("reserved", "Reserved"),
                ],
            ),
        ),
    ]



from django.db import models

from .. import exc
from .attribute import Attribute
from .resource import Resource


class Protocol(Resource):
    """
    Representation of a routing protocol
    """

    # Derive site from device; returns Site object (not site_id)
    site_source_field = "device"
    site_returns_object = True

    site = models.ForeignKey(
        "Site",
        db_index=True,
        blank=True,
        null=True,
        related_name="protocols",
        on_delete=models.PROTECT,
        verbose_name="Site",
        help_text=(
            "Unique ID of the Site this Protocol is under. If not set, this "
            "be inherited off of the device's site"
        ),
    )
    type = models.ForeignKey(
        "ProtocolType",
        db_index=True,
        related_name="protocols",
        on_delete=models.CASCADE,
        help_text="The type of this Protocol",
    )
    device = models.ForeignKey(
        "Device",
        db_index=True,
        null=False,
        related_name="protocols",
        on_delete=models.CASCADE,
        help_text="Device that this protocol is running on",
    )
    interface = models.ForeignKey(
        "Interface",
        db_index=True,
        blank=True,
        null=True,
        related_name="protocols",
        on_delete=models.CASCADE,
        help_text=(
            "Interface this protocol is running on. Either interface or"
            " circuit must be populated."
        ),
    )
    circuit = models.ForeignKey(
        "Circuit",
        db_index=True,
        blank=True,
        null=True,
        related_name="protocols",
        on_delete=models.CASCADE,
        help_text="Circuit that this protocol is running over.",
    )
    auth_string = models.CharField(
        max_length=255,
        default="",
        blank=True,
        verbose_name="Auth String",
        help_text="Authentication string (such as MD5 sum)",
    )
    description = models.CharField(
        max_length=255,
        default="",
        blank=True,
        help_text="Description for this Protocol",
    )

    def __str__(self):
        description = str(self.type)

        if self.circuit:
            description += " over %s" % self.circuit
        elif self.interface:
            description += " on %s" % self.interface
        else:
            description += " on %s" % self.device

        return description

    class Meta:
        ordering = ("device",)

    def clean_circuit(self, value):
        """Ensure at least one endpoint on the circuit is on this device"""
        if value and value.interface_for(self.device) is None:
            raise exc.ValidationError(
                {
                    "circuit": (
                        "At least one endpoint of the circuit must match the "
                        "device"
                    )
                }
            )

        return value

    def clean_interface(self, value):
        """
        Ensure that the interface is bound to the same device this Protocol is
        bound to.
        """
        if value and value.device != self.device:
            raise exc.ValidationError(
                {
                    "interface": (
                        "The interface must be on the same device that this"
                        " Protocol is on"
                    )
                }
            )

        return value

    def clean_type(self, value):
        """Ensure that ProtocolType matches our site."""
        if self.site != value.site:
            raise exc.ValidationError(
                {"type": "The type must be on the same site as this Protocol"}
            )

        return value

    def set_attributes(self, attributes, valid_attributes=None, partial=False):
        """
        Ensure that all attributes are set that are required by the set
        ProtocolType.
        """
        required = self.type.get_required_attributes()
        if valid_attributes is None:
            valid_attributes = Attribute.all_by_name(
                "Protocol", site=self.site
            )

        # Temporarily mark required attributes as ``required`` at run-time for
        # injecting required_attributes into validation.
        for r in required:
            if r in valid_attributes:
                valid_attributes[r].required = True

        return super().set_attributes(
            attributes, valid_attributes=valid_attributes, partial=partial
        )

    def clean_fields(self, exclude=None):
        self.site = self.clean_site(self.site)
        self.type = self.clean_type(self.type)
        self.interface = self.clean_interface(self.interface)
        self.circuit = self.clean_circuit(self.circuit)

    # TODO(jathan): type, device, interface, circuit need indexing. We might
    # consider caching these values ON the Protocol object similarly how we've
    # done it with other objects, so that the related lookups are only done on
    # update.
    def to_dict(self):
        return {
            "id": self.id,
            "site": self.site_id,
            "type": self.type.name,
            "device": self.device.hostname,
            "interface": self.interface and self.interface.name_slug,
            "circuit": self.circuit and self.circuit.name_slug,
            "description": self.description,
            "auth_string": self.auth_string,
            "attributes": self.get_attributes(),
        }



import json
import logging

from cryptography.fernet import Fernet, InvalidToken
from django.conf import settings
from django.contrib.auth.models import (
    AbstractBaseUser,
    BaseUserManager,
    PermissionsMixin,
)
from django.core.mail import send_mail
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .. import exc, util, validators
from .site import Site

log = logging.getLogger(__name__)


class EmailUserManager(BaseUserManager):
    """Custom manager for email-based User.

    This replaces the manager formerly provided by ``django-custom-user``.
    """

    def _create_user(
        self, email, password, is_staff, is_superuser, **extra_fields
    ):
        """
        Create and save a User with the given email and password.

        :param str email: user email
        :param str password: user password
        :param bool is_staff: whether user staff or not
        :param bool is_superuser: whether user admin or not
        :return User: user
        :raise ValueError: email is not set
        """
        now = timezone.now()
        if not email:
            raise ValueError("The given email must be set")
        email = self.normalize_email(email)
        is_active = extra_fields.pop("is_active", True)
        user = self.model(
            email=email,
            is_staff=is_staff,
            is_active=is_active,
            is_superuser=is_superuser,
            last_login=now,
            date_joined=now,
            **extra_fields,
        )
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, **extra_fields):
        """Create and save a regular User with the given email and password."""
        is_staff = extra_fields.pop("is_staff", False)
        return self._create_user(
            email, password, is_staff, False, **extra_fields
        )

    def create_superuser(self, email, password, **extra_fields):
        """Create and save a superuser with the given email and password."""
        return self._create_user(email, password, True, True, **extra_fields)


class User(AbstractBaseUser, PermissionsMixin):
    """A custom user object that utilizes email as the username.

    Replaces the former ``custom_user.models.AbstractEmailUser`` base class
    with an equivalent inline implementation based on ``AbstractBaseUser``
    and ``PermissionsMixin``.

    The following attributes are inherited from the superclasses:
        * password
        * last_login
        * is_superuser
    """

    email = models.EmailField(
        _("email address"), max_length=255, unique=True, db_index=True
    )
    is_staff = models.BooleanField(
        _("staff status"),
        default=False,
        help_text=_(
            "Designates whether the user can log into this admin site."
        ),
    )
    is_active = models.BooleanField(
        _("active"),
        default=True,
        help_text=_(
            "Designates whether this user should be treated as "
            "active. Unselect this instead of deleting accounts."
        ),
    )
    date_joined = models.DateTimeField(_("date joined"), default=timezone.now)
    secret_key = models.CharField(
        max_length=44,
        default=util.generate_secret_key,
        help_text="The user's secret_key used for API authentication.",
    )

    objects = EmailUserManager()

    USERNAME_FIELD = "email"
    REQUIRED_FIELDS = []

    class Meta:
        verbose_name = _("user")
        verbose_name_plural = _("users")

    def get_full_name(self):
        """Return the email."""
        return self.email

    def get_short_name(self):
        """Return the email."""
        return self.email

    def email_user(self, subject, message, from_email=None, **kwargs):
        """Send an email to this User."""
        send_mail(subject, message, from_email, [self.email], **kwargs)

    @property
    def username(self):
        return self.get_username()

    def get_permissions(self):
        permissions = []
        if self.is_staff or self.is_superuser:
            permissions.append("admin")
        sites = Site.objects.all()

        return {
            str(site.id): {
                "permissions": permissions,
                "site_id": site.id,
                "user_id": self.id,
            }
            for site in sites
        }

    def rotate_secret_key(self):
        self.secret_key = util.generate_secret_key()
        self.save()

    def generate_auth_token(self):
        """Serialize user data and encrypt token."""
        # Serialize data and utf-encode
        data = json.dumps({"email": self.email})

        # Encrypt data w/ servers's secret_key (it must be utf-encoded)
        f = Fernet(str(settings.SECRET_KEY))
        auth_token = f.encrypt(bytes(data, "utf-8"))
        return auth_token

    def verify_secret_key(self, secret_key):
        """Validate secret_key"""
        return secret_key == self.secret_key

    @classmethod
    def verify_auth_token(cls, email, auth_token, expiration=None):
        """Verify token and return a User object."""
        if expiration is None:
            expiration = settings.AUTH_TOKEN_EXPIRY

        # First we lookup the user by email
        query = cls.objects.filter(email=email)
        user = query.first()

        if user is None:
            log.debug("Invalid user when verifying token")
            raise exc.ValidationError(
                {"auth_token": "Invalid user when verifying token"}
            )
            # return None  # Invalid user

        # Decrypt auth_token w/ user's secret_key
        f = Fernet(str(settings.SECRET_KEY))
        try:
            decrypted_data = f.decrypt(
                bytes(auth_token, "utf-8"), ttl=expiration
            )
        except InvalidToken:
            log.debug("Invalid/expired auth_token when decrypting.")
            raise exc.ValidationError(
                {"auth_token": "Invalid/expired auth_token."}
            )
            # return None  # Invalid token

        # Deserialize data
        try:
            data = json.loads(decrypted_data)
        except ValueError:
            log.debug("Token could not be deserialized.")
            raise exc.ValidationError(
                {"auth_token": "Token could not be deserialized."}
            )
            # return None  # Valid token, but expired

        if email != data["email"]:
            log.debug("Invalid user when deserializing.")
            raise exc.ValidationError(
                {"auth_token": "Invalid user when deserializing."}
            )
            # return None  # User email did not match payload
        return user

    def clean_email(self, value):
        return validators.validate_email(value)

    def clean_fields(self, exclude=None):
        self.email = self.clean_email(self.email)

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)

    def to_dict(self, with_permissions=False, with_secret_key=False):
        out = [
            ("id", self.id),
            ("email", self.email),
        ]

        if with_secret_key:
            out.append(("secret_key", self.secret_key))

        if with_permissions:
            out.append(("permissions", self.get_permissions()))

        return dict(out)



"""
Project-wide utilities.
"""

import collections
import logging
import shlex

from cryptography.fernet import Fernet
from django.core.exceptions import FieldDoesNotExist

log = logging.getLogger(__name__)

_TRUTHY = set(["true", "yes", "on", "1", ""])

__all__ = (
    "SetQuery",
    "cidr_to_dict",
    "generate_secret_key",
    "generate_settings",
    "get_field_attr",
    "initialize_app",
    "main",
    "normalize_auth_header",
    "parse_set_query",
    "qpbool",
    "slugify",
    "slugify_interface",
)


def qpbool(arg):
    """
    Convert "truthy" strings into Booleans.

    >>> qpbool('true')
    True

    :param arg:
        Truthy string
    """
    return str(arg).lower() in _TRUTHY


def normalize_auth_header(header):
    """
    Normalize a header name into WSGI-compatible format.

    >>> normalize_auth_header('X-NSoT-Email')
    'HTTP_X_NSOT_EMAIL'

    :param header:
        Header name
    """
    return "HTTP_" + header.upper().replace("-", "_")


def generate_secret_key():
    """
    Return a secret key suitable for use w/ Fernet.

    >>> generate_secret_key()
    '1BpuqeM5d5pi-U2vIsqeQ8YnTrXRRUAfqV-hu6eQ5Gw='
    """
    key = Fernet.generate_key()
    return str(key.decode("utf-8"))


def get_field_attr(model, field_name, attr_name):
    """
    Return the specified attribute value from a model field.

    This is used for field overrides in API serializers to retrieve the
    ``verbose_name`` and ``help_text`` model field attributes so they don't
    need to be duplicated in code.

    Example::

        >>> get_field_attr(models.Interface, 'parent', 'help_text')
        'Unique ID of the parent Interface.'

    :param model:
        Model class

    :param field_name:
        Model field name

    :param attr_name:
        Model field attribute name
    """
    try:
        field = model._meta.get_field(field_name)
    except (AttributeError, FieldDoesNotExist):
        return ""
    else:
        return getattr(field, attr_name, "")


def cidr_to_dict(cidr):
    """
    Take a cidr and return it as a dictionary.

    >>> cidr_to_dict('192.168.0.0/16')
    {'network_address': '192.168.0.0', 'prefix_length': 16}

    :param cidr:
        IPv4/IPv6 CIDR string
    """
    from .. import validators

    cidr = validators.validate_cidr(cidr)
    return {
        "network_address": cidr.network_address,
        "prefix_length": cidr.prefixlen,
    }


def slugify(s):
    """
    Slugify a string.

    This works in a less-agressive manner than Django's slugify, which simply
    drops most drops most non-alphanumeric characters and lowercases the entire
    string. It would likely to cause uniqueness conflicts for things like
    interface names, such as Eth1/2/3 and Eth12/3, which would slugify to be
    the same.

    >>> slugify('switch-foo01:Ethernet1/2')
    'switch-foo01:Ethernet1_2'

    :param s:
        String to slugify
    """

    disallowed_chars = ["/"]
    replacement = "_"

    for char in disallowed_chars:
        s = s.replace(char, replacement)

    return s


def slugify_interface(
    interface=None, device_hostname=None, name=None, **kwargs
):
    """
    Return a slug (natural key) representation of an Interface.

    If ``interface`` is not provided, ``device_hostname`` and ``name`` are
    required. If all are provided, ``interface`` will take precedence.

    :param interface:
        Interface dict

    :param device_hostname:
        Device hostname

    :param name:
        Interface name

    :rtype: str
    """
    if not (interface or (device_hostname and name)):
        raise RuntimeError("Either interface or device_hostname/name required")

    if interface is None:
        interface = dict(device_hostname=device_hostname, name=name)

    slug = "{device_hostname}:{name}".format(**interface)
    return slug


#: Namedtuple for resultant items from ``parse_set_query()``
SetQuery = collections.namedtuple("SetQuery", "action name value")


def parse_set_query(query):
    """
    Parse a representation of set operations for attribute/value pairs into
    (action, name, value) and return a list of ``SetQuery`` objects.

    Computes left-to-right evaluation, where the first character indicates the
    set operation:

    + "+" indicates a union
    + "-" indicates a difference
    + no marker indicates an intersection

    For example::

        >>> parse_set_query('+owner=team-networking')
        [SetQuery(action='union', name='owner', value='team-networking')]
        >>> parse_set_query('foo=bar')
        [SetQuery(action='intersection', name='foo', value='bar')]
        >>> parse_set_query('foo=bar -owner=team-networking')
        [SetQuery(action='intersection', name='foo', value='bar'),
         SetQuery(action='difference', name='owner', value='team-networking')]

    :param query:
        Set query string
    """
    log.debug("Incoming query = %r" % (query,))

    if not isinstance(query, str):
        raise TypeError("Query must be a string.")

    queries = shlex.split(query)

    attributes = []
    for q in queries:
        if q.startswith("+"):
            action = "union"
            q = q[1:]
        elif q.startswith("-"):
            action = "difference"
            q = q[1:]
        else:
            action = "intersection"

        name, _, value = q.partition("=")
        attributes.append(SetQuery(action, name, value))

    log.debug("Outgoing attributes = %r" % (attributes,))
    return attributes


#: Configuration template emitted when a user runs ``nsot-server init``.
CONFIG_TEMPLATE = '''
"""
This configuration file is just Python code. You may override any global
defaults by specifying them here.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""
from nsot.conf.settings import *

import os.path

# Path where the config is found.
CONF_ROOT = os.path.dirname(__file__)

# A boolean that turns on/off debug mode. Never deploy a site into production
# with DEBUG turned on.
# Default: False
DEBUG = False

############
# Database #
############
# https://docs.djangoproject.com/en/dev/ref/settings/#databases
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(CONF_ROOT, 'nsot.sqlite3'),
        'USER': 'nsot',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}

###############
# Application #
###############

# The address on which the application will listen.
# Default: localhost
NSOT_HOST = 'localhost'

# The port on which the application will be accessed.
# Default: 8990
NSOT_PORT = 8990

# The number of gunicorn worker processes for handling requests.
# Default: 4
NSOT_NUM_WORKERS = 4

# Timeout in seconds before gunicorn workers are killed/restarted.
# Default: 30
NSOT_WORKER_TIMEOUT = 30

# If True, serve static files directly from the app.
# Default: True
SERVE_STATIC_FILES = True

############
# Security #
############

# A URL-safe base64-encoded 32-byte key. This must be kept secret. Anyone with
# this key is able to create and read messages. This key is used for
# encryption/decryption of sessions and auth tokens.
SECRET_KEY = %(secret_key)r

# Header to check for Authenticated Email. This is intended for use behind an
# authenticating reverse proxy.
USER_AUTH_HEADER = 'X-NSoT-Email'

# The age, in seconds, until an AuthToken granted by the API will expire.
# Default: 600
AUTH_TOKEN_EXPIRY = 600  # 10 minutes

# A list of strings representing the host/domain names that this Django site
# can serve. This is a security measure to prevent an attacker from poisoning
# caches and triggering password reset emails with links to malicious hosts by
# submitting requests with a fake HTTP Host header, which is possible even
# under many seemingly-safe web server configurations.
# https://docs.djangoproject.com/en/1.8/ref/settings/#allowed-hosts
ALLOWED_HOSTS = ['*']

# Whether new users created via header authentication (X-NSoT-Email) are
# automatically granted superuser privileges.
# Default: True
# NSOT_NEW_USERS_AS_SUPERUSER = True
'''


def generate_settings(config_template=None):
    """
    Used to emit a generated configuration from ``config_template``.

    :param config_template:
        Config template
    """
    if config_template is None:
        config_template = CONFIG_TEMPLATE.strip()

    secret_key = generate_secret_key()
    return config_template % dict(secret_key=secret_key)


def initialize_app(config):
    """
    Actions to be performed prior to creating the Application object.

    :param config:
        Config object
    """
    # This currently does nothing but it's here in case we need it.


def main():
    """CLI application used to manage NSoT."""
    from logan.runner import run_app

    run_app(
        project="nsot",
        default_config_path="~/.nsot/nsot.conf.py",
        default_settings="nsot.conf.settings",
        settings_initializer=generate_settings,
        settings_envvar="NSOT_CONF",
        initializer=initialize_app,
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from django.db import migrations, models
import django.db.models.deletion

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0024_network_state"),
    ]

    operations = [
        migrations.AddField(
            model_name="value",
            name="site",
            field=models.ForeignKey(
                related_name="values",
                on_delete=django.db.models.deletion.PROTECT,
                default=1,
                to="nsot.Site",
            ),
            preserve_default=False,
        ),
    ]



import logging

import django_filters
from django.db.models import Q

from .. import models
from ..util import qpbool

log = logging.getLogger(__name__)


class ChangeFilter(django_filters.rest_framework.FilterSet):
    """Filter for Change objects."""

    user = django_filters.NumberFilter(field_name="user")
    change_at__gte = django_filters.IsoDateTimeFilter(
        field_name="change_at", lookup_expr="gte"
    )
    change_at__lte = django_filters.IsoDateTimeFilter(
        field_name="change_at", lookup_expr="lte"
    )

    class Meta:
        model = models.Change
        fields = [
            "event",
            "resource_name",
            "resource_id",
            "user",
            "change_at__gte",
            "change_at__lte",
        ]


class AttributeFilter(django_filters.rest_framework.FilterSet):
    required = django_filters.BooleanFilter()
    display = django_filters.BooleanFilter()
    multi = django_filters.BooleanFilter()
    has_dependencies = django_filters.BooleanFilter(
        method="filter_has_dependencies",
        help_text="Filter attributes that have dependencies (true) or don't (false).",
    )

    def filter_has_dependencies(self, queryset, name, value):
        if value:
            return queryset.filter(depends_on__isnull=False).distinct()
        return queryset.filter(depends_on__isnull=True)

    class Meta:
        model = models.Attribute
        fields = ["name", "resource_name", "required", "display", "multi"]


class ResourceFilter(django_filters.rest_framework.FilterSet):
    """Attribute-aware filtering for Resource objects."""

    attributes = django_filters.CharFilter(method="filter_attributes")
    expired = django_filters.BooleanFilter(
        method="filter_expired",
        help_text="Filter by expiration status (true=expired, false=not expired).",
    )
    expires_before = django_filters.IsoDateTimeFilter(
        field_name="expires_at",
        lookup_expr="lte",
        help_text="Resources expiring on or before this ISO timestamp.",
    )
    expires_after = django_filters.IsoDateTimeFilter(
        field_name="expires_at",
        lookup_expr="gte",
        help_text="Resources expiring on or after this ISO timestamp.",
    )

    def filter_expired(self, queryset, name, value):
        """Delegate to the queryset's ``.expired()`` method."""
        return queryset.expired(expired=value)

    def filter_attributes(self, queryset, name, value):
        """
        Reads 'attributes' from query params and joins them together as an
        intersection set query.
        """
        attributes = self.data.getlist("attributes", [])
        resource_name = queryset.model.__name__

        # Iterate the attributes and try to look them up as if they are k=v
        # and naively do an intersection query.
        log.debug("GOT ATTRIBUTES: %r", attributes)

        for attribute in attributes:
            name, _, value = attribute.partition("=")
            # Retrieve next set of objects using the same arguments as the
            # initial query.
            next_set = Q(
                id__in=models.Value.objects.filter(
                    name=name, value=value, resource_name=resource_name
                ).values_list("resource_id", flat=True)
            )
            queryset = queryset.filter(next_set)

        return queryset


class DeviceFilter(ResourceFilter):
    """Filter for Device objects."""

    class Meta:
        model = models.Device
        fields = [
            "hostname",
            "attributes",
            "expired",
            "expires_before",
            "expires_after",
        ]


class NetworkFilter(ResourceFilter):
    """Filter for Network objects."""

    include_networks = django_filters.BooleanFilter(
        method="filter_include_networks"
    )
    include_ips = django_filters.BooleanFilter(method="filter_include_ips")
    cidr = django_filters.CharFilter(method="filter_cidr")
    root_only = django_filters.BooleanFilter(method="filter_root_only")
    network_address = django_filters.CharFilter()  # Override type

    class Meta:
        model = models.Network
        fields = [
            "include_networks",
            "include_ips",
            "root_only",
            "cidr",
            "network_address",
            "prefix_length",
            "ip_version",
            "state",
            "attributes",
            "expired",
            "expires_before",
            "expires_after",
        ]

    def filter_include_networks(self, queryset, name, value):
        """Converts ``include_networks`` to queryset filters."""
        include_ips = qpbool(self.form.cleaned_data["include_ips"])
        include_networks = qpbool(value)

        if not all([include_networks, include_ips]):
            if include_networks:
                return queryset.filter(is_ip=False)
            return queryset.exclude(is_ip=False)

        return queryset

    def filter_include_ips(self, queryset, name, value):
        """Converts ``include_ips`` to queryset filters."""
        include_ips = qpbool(value)
        include_networks = qpbool(self.form.cleaned_data["include_networks"])

        if not all([include_networks, include_ips]):
            if include_ips:
                return queryset.filter(is_ip=True)
            return queryset.exclude(is_ip=True)

        return queryset

    def filter_cidr(self, queryset, name, value):
        """Converts ``cidr`` to network/prefix filter."""
        if value:
            network_address, _, prefix_length = value.partition("/")
        else:
            return queryset

        return queryset.filter(
            network_address=network_address, prefix_length=prefix_length
        )

    def filter_root_only(self, queryset, name, value):
        """Converts ``root_only`` to null parent filter."""
        if qpbool(value):
            return queryset.filter(parent=None)
        return queryset


class InterfaceFilter(ResourceFilter):
    """
    Filter for Interface objects.

    Includes a custom override for filtering on mac_address because this is not
    a Django built-in field.
    """

    mac_address = django_filters.CharFilter(method="filter_mac_address")

    class Meta:
        model = models.Interface
        # TODO: Remove `device__hostname` in a future release after
        #       updating pynsot to use `device_hostname`.
        fields = [
            "device",
            "device__hostname",
            "name",
            "speed",
            "mtu",
            "type",
            "mac_address",
            "description",
            "parent_id",
            "attributes",
            "device_hostname",
            "expired",
            "expires_before",
            "expires_after",
        ]

    def filter_mac_address(self, queryset, name, value):
        """
        Overloads queryset filtering to use built-in.

        Doesn't work by default because MACAddressField is not a Django
        built-in field type.
        """
        return queryset.filter(mac_address=value)


class CircuitFilter(ResourceFilter):
    """Filter for Circuit objects."""

    endpoint_a = django_filters.CharFilter(method="filter_endpoint_a")
    endpoint_z = django_filters.CharFilter(method="filter_endpoint_z")
    device_hostname = django_filters.CharFilter(
        method="filter_device_hostname"
    )

    class Meta:
        model = models.Circuit
        fields = [
            "endpoint_a",
            "endpoint_z",
            "name",
            "device_hostname",
            "attributes",
            "expired",
            "expires_before",
            "expires_after",
        ]

    # FIXME(jathan): The copy/pasted methods can be ripped out once we upgrade
    # filters in support of the V2 API. For now this is quicker and easier.
    def filter_endpoint_a(self, queryset, name, value):
        """Overload to use natural key."""
        if isinstance(value, int):
            value = str(value)

        if value.isdigit():
            return queryset.filter(endpoint_a=value)
        return queryset.filter(endpoint_a__name_slug=value)

    def filter_endpoint_z(self, queryset, name, value):
        """Overload to use natural key."""
        if isinstance(value, int):
            value = str(value)

        if value.isdigit():
            return queryset.filter(endpoint_z=value)
        return queryset.filter(endpoint_z__name_slug=value)

    def filter_device_hostname(self, queryset, name, value):
        """Filter circuits by device hostname on either endpoint."""
        return queryset.filter(
            Q(endpoint_a__device__hostname=value)
            | Q(endpoint_z__device__hostname=value)
        )


class ProtocolTypeFilter(django_filters.rest_framework.FilterSet):
    """Filter for ProtocolType (non-resource) objects."""

    class Meta:
        model = models.ProtocolType
        fields = ["name", "description"]


class ProtocolFilter(ResourceFilter):
    """Filter for Protocol objects."""

    device = django_filters.CharFilter(method="filter_device")
    type = django_filters.CharFilter(method="filter_type")
    interface = django_filters.CharFilter(method="filter_interface")
    circuit = django_filters.CharFilter(method="filter_circuit")

    class Meta:
        model = models.Protocol
        fields = [
            "device",
            "type",
            "interface",
            "circuit",
            "description",
            "expired",
            "expires_before",
            "expires_after",
        ]

    def filter_device(self, queryset, name, value):
        """Overload to use natural key."""
        if isinstance(value, int):
            value = str(value)

        if value.isdigit():
            return queryset.filter(device=value)
        return queryset.filter(device__hostname=value)

    def filter_type(self, queryset, name, value):
        """Overload to use natural key."""
        if isinstance(value, int):
            value = str(value)

        if value.isdigit():
            return queryset.filter(type=value)
        return queryset.filter(type__name=value)

    def filter_interface(self, queryset, name, value):
        """Overload to use natural key."""
        if isinstance(value, int):
            value = str(value)

        if value.isdigit():
            return queryset.filter(interface=value)
        return queryset.filter(interface__name_slug=value)

    def filter_circuit(self, queryset, name, value):
        """Overload to use natural key."""
        if isinstance(value, int):
            value = str(value)

        if value.isdigit():
            return queryset.filter(circuit=value)
        return queryset.filter(circuit__name_slug=value)



# Convert index_together to Meta.indexes for Django 5.x compatibility.
#
# Models where index_together duplicated unique_together (attribute,
# assignment, device, network, interface's device+name) don't need new
# indexes — the unique constraint already provides one.
#
# Models with standalone indexes (value, change, interface's
# device_hostname+name) get fresh named indexes via AddIndex.
# We use AddIndex instead of RenameIndex because SQLite table
# rebuilds in prior migrations (AlterField) can lose unnamed
# index_together indexes.

import nsot.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0039_django30_compat"),
    ]

    operations = [
        # Clear ALL index_together from state.  For models that overlap
        # with unique_together, no DB change is needed.  For standalone
        # indexes, we create fresh named indexes below.
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterIndexTogether(
                    name="attribute", index_together=None,
                ),
                migrations.AlterIndexTogether(
                    name="assignment", index_together=None,
                ),
                migrations.AlterIndexTogether(
                    name="change", index_together=None,
                ),
                migrations.AlterIndexTogether(
                    name="device", index_together=None,
                ),
                migrations.AlterIndexTogether(
                    name="interface", index_together=None,
                ),
                migrations.AlterIndexTogether(
                    name="network", index_together=None,
                ),
                migrations.AlterIndexTogether(
                    name="value", index_together=None,
                ),
            ],
            database_operations=[],
        ),
        # Create named indexes for standalone index_together entries.
        # Drop any existing indexes first to avoid duplicates on PostgreSQL/MySQL.
        # On SQLite, these may not exist due to table rebuilds, hence IF EXISTS.
        #
        # Interface: device_hostname + name (from 0027)
        migrations.RunSQL(
            sql="DROP INDEX IF EXISTS nsot_interface_device_hostname_name",
            reverse_sql=migrations.RunSQL.noop,
        ),
        migrations.AddIndex(
            model_name="interface",
            index=models.Index(
                fields=["device_hostname", "name"],
                name="nsot_interf_device__f01451_idx",
            ),
        ),
        # Value: (name, value, resource_name) - explicitly named in 0021
        migrations.RunSQL(
            sql="DROP INDEX IF EXISTS nsot_val_name_val_rn",
            reverse_sql=migrations.RunSQL.noop,
        ),
        migrations.AddIndex(
            model_name="value",
            index=models.Index(
                fields=["name", "value", "resource_name"],
                name="nsot_value_name_0c66ac_idx",
            ),
        ),
        # Value: (resource_name, resource_id) - from 0023
        migrations.RunSQL(
            sql="DROP INDEX IF EXISTS nsot_value_resource_name_resource_id",
            reverse_sql=migrations.RunSQL.noop,
        ),
        migrations.AddIndex(
            model_name="value",
            index=models.Index(
                fields=["resource_name", "resource_id"],
                name="nsot_value_resourc_62e5c1_idx",
            ),
        ),
        # Change: (resource_name, resource_id) and (resource_name, event) - from 0032
        migrations.RunSQL(
            sql="DROP INDEX IF EXISTS nsot_change_resource_name_resource_id",
            reverse_sql=migrations.RunSQL.noop,
        ),
        migrations.AddIndex(
            model_name="change",
            index=models.Index(
                fields=["resource_name", "resource_id"],
                name="nsot_change_resourc_fa9b1d_idx",
            ),
        ),
        migrations.RunSQL(
            sql="DROP INDEX IF EXISTS nsot_change_resource_name_event",
            reverse_sql=migrations.RunSQL.noop,
        ),
        migrations.AddIndex(
            model_name="change",
            index=models.Index(
                fields=["resource_name", "event"],
                name="nsot_change_resourc_9023b9_idx",
            ),
        ),
        # mac_address field type change detected by makemigrations.
        migrations.AlterField(
            model_name="interface",
            name="mac_address",
            field=nsot.fields.MACAddressField(
                blank=True,
                db_index=True,
                default=0,
                help_text="If not provided, defaults to 00:00:00:00:00:00.",
                integer=True,
                null=True,
                verbose_name="MAC Address",
            ),
        ),
    ]



"""Command to generate a secret key."""

from cryptography.fernet import Fernet

from nsot.util.commands import NsotCommand


class Command(NsotCommand):
    help = (
        "Generate a URL-safe base64-encoded 32-byte key for use in "
        "settings.SECRET_KEY."
    )

    def handle(self, **options):
        print(Fernet.generate_key())



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django.db.models.deletion
import django_extensions.db.fields.json
import nsot.fields

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0014_auto_20151002_1653"),
    ]

    operations = [
        ##########
        # Device #
        ##########
        # Create temp Device object
        migrations.CreateModel(
            name="Device_temp",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "_attributes_cache",
                    django_extensions.db.fields.json.JSONField(blank=True),
                ),
                ("hostname", models.CharField(max_length=255, db_index=True)),
                (
                    "site",
                    models.ForeignKey(
                        related_name="devices",
                        on_delete=django.db.models.deletion.PROTECT,
                        to="nsot.Site",
                    ),
                ),
            ],
        ),
        # self._attributes_cache = self.resource_ptr._attributes
        migrations.AddField(
            model_name="device",
            name="_attributes_cache",
            field=django_extensions.db.fields.json.JSONField(blank=True),
        ),
        # self.id = self.resource_ptr_id
        migrations.AddField(
            model_name="device",
            name="new_id",
            field=models.IntegerField(default=0, verbose_name="ID"),
            preserve_default=False,
        ),
        ###########
        # Network #
        ###########
        migrations.CreateModel(
            name="Network_temp",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "_attributes_cache",
                    django_extensions.db.fields.json.JSONField(blank=True),
                ),
                (
                    "network_address",
                    nsot.fields.BinaryIPAddressField(max_length=16, db_index=True),
                ),
                (
                    "broadcast_address",
                    nsot.fields.BinaryIPAddressField(max_length=16, db_index=True),
                ),
                ("prefix_length", models.IntegerField(db_index=True)),
                (
                    "ip_version",
                    models.CharField(
                        db_index=True, max_length=1, choices=[("4", "4"), ("6", "6")]
                    ),
                ),
                ("is_ip", models.BooleanField(default=False, db_index=True)),
                (
                    "parent",
                    models.ForeignKey(
                        related_name="children",
                        on_delete=django.db.models.deletion.PROTECT,
                        default=None,
                        blank=True,
                        to="nsot.Network_temp",
                        null=True,
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        related_name="networks",
                        on_delete=django.db.models.deletion.PROTECT,
                        to="nsot.Site",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="network",
            name="_attributes_cache",
            field=django_extensions.db.fields.json.JSONField(blank=True),
        ),
        migrations.AddField(
            model_name="network",
            name="new_id",
            field=models.IntegerField(default=0, verbose_name="ID"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="network",
            name="new_parent_id",
            field=models.IntegerField(null=True, verbose_name="Parent ID"),
            preserve_default=False,
        ),
        ##################
        # Assignment 1/2 #
        ##################
        migrations.CreateModel(
            name="Assignment_temp",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("address", models.ForeignKey(to="nsot.Network_temp", on_delete=django.db.models.deletion.CASCADE)),
                # ('interface', models.ForeignKey(to='nsot.Interface_temp')),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        #############
        # Interface #
        #############
        migrations.CreateModel(
            name="Interface_temp",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "_attributes_cache",
                    django_extensions.db.fields.json.JSONField(blank=True),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="The name of the interface as it appears on the Device.",
                        max_length=255,
                        db_index=True,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        default="",
                        help_text="A brief yet helpful description.",
                        max_length=255,
                        blank=True,
                    ),
                ),
                (
                    "type",
                    models.IntegerField(
                        default=6,
                        help_text="If not provided, defaults to 'ethernet'.",
                        db_index=True,
                        verbose_name="Interface Type",
                        choices=[
                            (6, b"ethernet"),
                            (1, b"other"),
                            (135, b"l2vlan"),
                            (136, b"l3vlan"),
                            (161, b"lag"),
                            (24, b"loopback"),
                            (150, b"mpls"),
                            (53, b"prop_virtual"),
                            (131, b"tunnel"),
                        ],
                    ),
                ),
                (
                    "mac_address",
                    nsot.fields.MACAddressField(
                        default=0,
                        blank=True,
                        help_text="If not provided, defaults to 00:00:00:00:00:00.",
                        integer=True,
                        null=True,
                        verbose_name="MAC Address",
                        db_index=True,
                    ),
                ),
                (
                    "speed",
                    models.IntegerField(
                        default=10000,
                        help_text="Integer of Mbps of interface (e.g. 20000 for 20 Gbps). If not provided, defaults to 10000.",
                        db_index=True,
                        blank=True,
                    ),
                ),
                (
                    "addresses",
                    models.ManyToManyField(
                        related_name="addresses",
                        through="nsot.Assignment_temp",
                        to="nsot.Network_temp",
                        db_index=True,
                    ),
                ),
                (
                    "device",
                    models.ForeignKey(
                        related_name="interfaces",
                        verbose_name="Device",
                        to="nsot.Device_temp",
                        on_delete=django.db.models.deletion.CASCADE,
                        help_text="Unique ID of the connected Device.",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        related_name="children",
                        on_delete=django.db.models.deletion.PROTECT,
                        default=None,
                        to="nsot.Interface_temp",
                        blank=True,
                        help_text="Unique ID of the parent Interface.",
                        null=True,
                        verbose_name="Parent",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        related_name="interfaces",
                        on_delete=django.db.models.deletion.PROTECT,
                        to="nsot.Site",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="interface",
            name="_attributes_cache",
            field=django_extensions.db.fields.json.JSONField(blank=True),
        ),
        migrations.AddField(
            model_name="interface",
            name="new_id",
            field=models.IntegerField(default=0, verbose_name="ID"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="interface",
            name="new_parent_id",
            field=models.IntegerField(null=True, verbose_name="Parent ID"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="device",
            name="new_device_id",
            field=models.IntegerField(default=0, verbose_name="Device ID"),
            preserve_default=False,
        ),
        ##################
        # Assignment 2/2 #
        ##################
        migrations.AddField(
            model_name="assignment_temp",
            name="interface",
            field=models.ForeignKey(to="nsot.Interface_temp", on_delete=django.db.models.deletion.CASCADE),
        ),
        migrations.AlterUniqueTogether(
            name="interface_temp",
            unique_together=set([("device", "name")]),
        ),
        migrations.AlterIndexTogether(
            name="interface_temp",
            index_together=set([("device", "name")]),
        ),
        migrations.AlterUniqueTogether(
            name="assignment_temp",
            unique_together=set([("address", "interface")]),
        ),
        migrations.AlterIndexTogether(
            name="assignment_temp",
            index_together=set([("address", "interface")]),
        ),
        #########
        # Value #
        #########
        # self.attribute.name => self.name
        migrations.AddField(
            model_name="value",
            name="name",
            field=models.CharField(max_length=64, verbose_name="Name", blank=True),
        ),
        # self.resource_id => self.new_resource_id
        migrations.AddField(
            model_name="value",
            name="new_resource_id",
            field=models.IntegerField(default=1, verbose_name="Resource ID"),
            preserve_default=False,
        ),
        # self.resource_name = self.resource.__class__.__name__
        migrations.AddField(
            model_name="value",
            name="resource_name",
            field=models.CharField(
                default="BOGUS",
                max_length=20,
                verbose_name="Resource Type",
                db_index=True,
                choices=[
                    ("Device", "Device"),
                    ("Attribute", "Attribute"),
                    ("Interface", "Interface"),
                    ("Site", "Site"),
                    ("Network", "Network"),
                ],
            ),
            preserve_default=False,
        ),
        ##########
        # Change #
        ##########
        migrations.AlterField(
            model_name="change",
            name="resource_name",
            field=models.CharField(
                db_index=True,
                max_length=20,
                verbose_name="Resource Type",
                choices=[
                    ("Device", "Device"),
                    ("Attribute", "Attribute"),
                    ("Interface", "Interface"),
                    ("Site", "Site"),
                    ("Network", "Network"),
                ],
            ),
        ),
    ]



from django.core.exceptions import ValidationError
from rest_framework import status
from rest_framework.mixins import CreateModelMixin
from rest_framework.response import Response

__all__ = [
    "BulkCreateModelMixin",
    "BulkDestroyModelMixin",
    "BulkUpdateModelMixin",
]


class BulkCreateModelMixin(CreateModelMixin):
    """
    Either create a single or many model instances in bulk by using the
    Serializers ``many=True`` ability from Django REST >= 2.2.5.

    .. note::
        This mixin uses the same method to create model instances
        as ``CreateModelMixin`` because both non-bulk and bulk
        requests will use ``POST`` request method.
    """

    def create(self, request, *args, **kwargs):
        bulk = isinstance(request.DATA, list)

        if not bulk:
            return super(BulkCreateModelMixin, self).create(
                request, *args, **kwargs
            )

        else:
            serializer = self.get_serializer(data=request.DATA, many=True)
            if serializer.is_valid():
                for obj in serializer.object:
                    self.pre_save(obj)
                self.object = serializer.save(force_insert=True)
                for obj in self.object:
                    self.post_save(obj, created=True)
                return Response(
                    serializer.data, status=status.HTTP_201_CREATED
                )

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class BulkUpdateModelMixin(object):
    """
    Update model instances in bulk by using the Serializers
    ``many=True`` ability from Django REST >= 2.2.5.
    """

    def get_object(self, queryset=None):
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        if any(
            (
                lookup_url_kwarg in self.kwargs,
                self.pk_url_kwarg in self.kwargs,
                self.slug_url_kwarg in self.kwargs,
            )
        ):
            return super(BulkUpdateModelMixin, self).get_object(queryset)

        # If the lookup_url_kwarg (or other deprecated variations)
        # are not present, get_object() is most likely called
        # as part of metadata() which by default simply checks
        # for object permissions and raises permission denied if necessary.
        # Here we don't need to check for general permissions
        # and can simply return None since general permissions
        # are checked in initial() which always gets executed
        # before any of the API actions (e.g. create, update, etc)
        return

    def bulk_update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)

        # restrict the update to the filtered queryset
        serializer = self.get_serializer(
            self.filter_queryset(self.get_queryset()),
            data=request.DATA,
            many=True,
            partial=partial,
        )

        if serializer.is_valid():
            try:
                for obj in serializer.object:
                    self.pre_save(obj)
            except ValidationError as err:
                # full_clean on model instances may be called in pre_save
                # so we have to handle eventual errors.
                return Response(
                    err.message_dict, status=status.HTTP_400_BAD_REQUEST
                )
            self.object = serializer.save(force_update=True)
            for obj in self.object:
                self.post_save(obj, created=False)
            return Response(serializer.data, status=status.HTTP_200_OK)

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def partial_bulk_update(self, request, *args, **kwargs):
        kwargs["partial"] = True
        return self.bulk_update(request, *args, **kwargs)


class BulkDestroyModelMixin(object):
    """
    Destroy model instances.
    """

    def allow_bulk_destroy(self, qs, filtered):
        """
        Hook to ensure that the bulk destroy should be allowed.

        By default this checks that the destroy is only applied to
        filtered querysets.
        """
        return qs is not filtered

    def bulk_destroy(self, request, *args, **kwargs):
        qs = self.get_queryset()
        filtered = self.filter_queryset(qs)
        if not self.allow_bulk_destroy(qs, filtered):
            return Response(status=status.HTTP_400_BAD_REQUEST)

        for obj in filtered:
            self.pre_delete(obj)
            obj.delete()
            self.post_delete(obj)
        return Response(status=status.HTTP_204_NO_CONTENT)



import difflib
import json
from calendar import timegm

from django.conf import settings
from django.db import models

from .. import exc, fields
from . import constants
from .site import Site


class Change(models.Model):
    """Record of all changes in NSoT."""

    site = models.ForeignKey(
        "Site",
        db_index=True,
        related_name="changes",
        verbose_name="Site",
        on_delete=models.CASCADE,
        help_text="Unique ID of the Site this Change is under.",
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="changes",
        db_index=True,
        on_delete=models.CASCADE,
        help_text="The User that initiated this Change.",
    )
    change_at = models.DateTimeField(
        auto_now_add=True,
        db_index=True,
        null=False,
        help_text="The timestamp of this Change.",
    )
    event = models.CharField(
        max_length=10,
        null=False,
        choices=constants.EVENT_CHOICES,
        help_text="The type of event this Change represents.",
    )
    resource_id = models.IntegerField(
        "Resource ID",
        null=False,
        help_text="The unique ID of the Resource for this Change.",
    )
    resource_name = models.CharField(
        "Resource Type",
        max_length=20,
        null=False,
        db_index=True,
        choices=constants.CHANGE_RESOURCE_CHOICES,
        help_text="The name of the Resource for this Change.",
    )
    _resource = fields.JSONField(
        "Resource",
        null=False,
        blank=True,
        help_text="Local cache of the changed Resource. (Internal use only)",
    )

    def __init__(self, *args, **kwargs):
        self._obj = kwargs.pop("obj", None)
        super().__init__(*args, **kwargs)

    class Meta:
        get_latest_by = "change_at"
        indexes = [
            models.Index(fields=["resource_name", "resource_id"]),
            models.Index(fields=["resource_name", "event"]),
        ]

    def __str__(self):
        return "%s %s(%s)" % (self.event, self.resource_name, self.resource_id)

    @property
    def resource(self):
        return self._resource

    def get_change_at(self):
        return timegm(self.change_at.timetuple())

    get_change_at.short_description = "Change At"

    @classmethod
    def get_serializer_for_resource(cls, resource_name):
        from ..api import serializers

        serializer_class = resource_name + "Serializer"
        return getattr(serializers, serializer_class)

    def clean_event(self, value):
        if value not in constants.CHANGE_EVENTS:
            raise exc.ValidationError("Invalid change event: %r." % value)
        return value

    def clean_resource_name(self, value):
        if value not in constants.VALID_CHANGE_RESOURCES:
            raise exc.ValidationError("Invalid resource name: %r." % value)
        return value

    def clean_site(self, obj):
        """value in this case is an instance of a model object."""

        # Site doesn't have an id to itself, so if obj is a Site, use it.
        # Otherwise get the value of the `.site`
        return obj if isinstance(obj, Site) else obj.site

    def clean_fields(self, exclude=None):
        """This will populate the change fields from the incoming object."""
        obj = self._obj
        if obj is None:
            return

        self.event = self.clean_event(self.event)
        self.resource_name = self.clean_resource_name(obj.__class__.__name__)
        self.resource_id = obj.id
        self.site = self.clean_site(obj)

        serializer_class = self.get_serializer_for_resource(self.resource_name)
        serializer = serializer_class(obj)
        self._resource = serializer.data

    def save(self, *args, **kwargs):
        self.full_clean()  # First validate fields are correct
        super().save(*args, **kwargs)

    def to_dict(self):
        resource = None
        if self.resource is not None:
            resource = self.resource

        return {
            "id": self.id,
            "site": self.site.to_dict(),
            "user": self.user.to_dict(),
            "change_at": timegm(self.change_at.timetuple()),
            "event": self.event,
            "resource_name": self.resource_name,
            "resource_id": self.resource_id,
            "resource": resource,
        }

    def _get_previous_change(self):
        """Return the most recent prior Change for the same resource, or
        None."""
        return (
            Change.objects.filter(
                change_at__lt=self.change_at,
                resource_id=self.resource_id,
                resource_name=self.resource_name,
            )
            .order_by("-change_at")
            .first()
        )

    @property
    def resource_diff(self):
        """Return a dict of changed fields with old/new values."""
        if self.event == "Create":
            return {
                k: {"old": None, "new": v} for k, v in self._resource.items()
            }

        if self.event == "Delete":
            return {
                k: {"old": v, "new": None} for k, v in self._resource.items()
            }

        # Update — show only changed fields
        prev = self._get_previous_change()
        if prev is None:
            return {
                k: {"old": None, "new": v} for k, v in self._resource.items()
            }

        diff = {}
        all_keys = set(prev._resource) | set(self._resource)
        for key in all_keys:
            old_val = prev._resource.get(key)
            new_val = self._resource.get(key)
            if old_val != new_val:
                diff[key] = {"old": old_val, "new": new_val}
        return diff

    @property
    def diff(self):
        """Return a text diff between previous and current resource state."""
        if self.event == "Create":
            old = ""
        elif self.event == "Delete":
            old = json.dumps(self._resource, indent=2, sort_keys=True)
        else:
            prev = self._get_previous_change()
            old = (
                json.dumps(prev._resource, indent=2, sort_keys=True)
                if prev
                else ""
            )

        if self.event == "Delete":
            current = ""
        else:
            current = json.dumps(self._resource, indent=2, sort_keys=True)

        return "\n".join(difflib.ndiff(old.splitlines(), current.splitlines()))



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django.db.models.deletion
import django_extensions.db.fields.json

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0020_move_value_data"),
    ]
    atomic = False

    operations = [
        #########
        # Value #
        #########
        # Fix constraints
        migrations.AlterUniqueTogether(
            name="value",
            unique_together=set(
                [("name", "value", "resource_name", "new_resource_id")]
            ),
        ),
        # Wrapped in SeparateDatabaseAndState because Django 5.x is strict
        # about dropping index_together indexes that overlap with
        # unique_together (the old index from 0001 may not exist as a
        # separate non-unique index). We update state and only CREATE
        # the new index.
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterIndexTogether(
                    name="value",
                    index_together=set([("name", "value", "resource_name")]),
                ),
            ],
            database_operations=[
                migrations.RunSQL(
                    sql="CREATE INDEX IF NOT EXISTS nsot_val_name_val_rn ON nsot_value (name, value, resource_name)",
                    reverse_sql="DROP INDEX IF EXISTS nsot_val_name_val_rn",
                ),
            ],
        ),
        # Delete the resource fk
        migrations.RemoveField(
            model_name="value",
            name="resource",
        ),
        # Rename resource_id
        migrations.RenameField(
            model_name="value",
            old_name="new_resource_id",
            new_name="resource_id",
        ),
        #############
        # Resources #
        #############
        # Delete the now-legacy Resource-based multiple inheritance models.
        migrations.DeleteModel(
            name="Assignment",
        ),
        migrations.DeleteModel(
            name="Interface",
        ),
        migrations.DeleteModel(
            name="Network",
        ),
        migrations.DeleteModel(
            name="Device",
        ),
        # Delete the Resource model altogether
        migrations.DeleteModel(
            name="Resource",
        ),
        #################
        # Rename models #
        #################
        # Start renaming the new Resource-abstract concrete models.
        # Network & Network relations
        migrations.RenameModel(
            old_name="Network_temp",
            new_name="Network",
        ),
        migrations.AlterField(
            model_name="network",
            name="parent",
            field=models.ForeignKey(
                related_name="children",
                on_delete=django.db.models.deletion.PROTECT,
                default=None,
                blank=True,
                to="nsot.Network",
                null=True,
            ),
        ),
        # Device
        migrations.RenameModel(
            old_name="Device_temp",
            new_name="Device",
        ),
        # Interface & Interface relations
        migrations.RenameModel(
            old_name="Interface_temp",
            new_name="Interface",
        ),
        ##############
        # Assignment #
        ##############
        # Because Assignments is used as m2m-through on Interface, we have to
        # create the model anew.
        migrations.CreateModel(
            name="Assignment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "address",
                    models.ForeignKey(
                        related_name="assignments",
                        to="nsot.Network",
                        on_delete=django.db.models.deletion.CASCADE
                    ),
                ),
                (
                    "interface",
                    models.ForeignKey(
                        related_name="assignments",
                        to="nsot.Interface",
                        db_index=True,
                        on_delete=django.db.models.deletion.CASCADE,
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        # Interface => m2m throgh Assignment => Network
        migrations.AlterField(
            model_name="interface",
            name="addresses",
            field=models.ManyToManyField(
                related_name="addresses",
                through="nsot.Assignment",
                to="nsot.Network",
                db_index=True,
            ),
        ),
        # Ditch the intermediate temp Assignment model
        migrations.DeleteModel(
            name="Assignment_temp",
        ),
        # Interface relations
        migrations.AlterField(
            model_name="interface",
            name="device",
            field=models.ForeignKey(
                related_name="interfaces",
                verbose_name="Device",
                to="nsot.Device",
                on_delete=django.db.models.deletion.CASCADE,
                help_text="Unique ID of the connected Device.",
            ),
        ),
        migrations.AlterField(
            model_name="interface",
            name="parent",
            field=models.ForeignKey(
                related_name="children",
                on_delete=django.db.models.deletion.PROTECT,
                default=None,
                to="nsot.Interface",
                blank=True,
                help_text="Unique ID of the parent Interface.",
                null=True,
                verbose_name="Parent",
            ),
        ),
        ###############
        # Constraints #
        ###############
        migrations.AlterUniqueTogether(
            name="assignment",
            unique_together=set([("address", "interface")]),
        ),
        migrations.AlterUniqueTogether(
            name="device",
            unique_together=set([("site", "hostname")]),
        ),
        migrations.AlterUniqueTogether(
            name="network",
            unique_together=set(
                [("site", "ip_version", "network_address", "prefix_length")]
            ),
        ),
        migrations.AlterIndexTogether(
            name="assignment",
            index_together=set([("address", "interface")]),
        ),
        migrations.AlterIndexTogether(
            name="device",
            index_together=set([("site", "hostname")]),
        ),
        migrations.AlterIndexTogether(
            name="network",
            index_together=set(
                [("site", "ip_version", "network_address", "prefix_length")]
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations, models

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0026_model_field_verbose_names"),
    ]

    operations = [
        migrations.AddField(
            model_name="interface",
            name="device_hostname",
            field=models.CharField(
                help_text="The hostname of the Device to which the interface is bound. (Internal use only)",
                max_length=255,
                db_index=True,
                blank=True,
                editable=False,
            ),
        ),
        migrations.AlterIndexTogether(
            name="interface",
            index_together=set([("device_hostname", "name"), ("device", "name")]),
        ),
    ]



from django.db import models

from .. import exc
from . import constants
from .attribute import Attribute


class Value(models.Model):
    """Represents a value for an attribute attached to a Resource."""

    attribute = models.ForeignKey(
        "Attribute",
        related_name="values",
        db_index=True,
        on_delete=models.PROTECT,
        help_text="The Attribute to which this Value is assigned.",
    )
    value = models.CharField(
        max_length=255,
        null=False,
        blank=True,
        db_index=True,
        help_text="The Attribute value.",
    )
    resource_id = models.IntegerField(
        "Resource ID",
        null=False,
        help_text="The unique ID of the Resource to which the Value is bound.",
    )
    resource_name = models.CharField(
        "Resource Type",
        max_length=20,
        null=False,
        db_index=True,
        choices=constants.RESOURCE_CHOICES,
        help_text="The name of the Resource type to which the Value is bound.",
    )
    name = models.CharField(
        "Name",
        max_length=64,
        null=False,
        blank=True,
        help_text=(
            "The name of the Attribute to which the Value is bound. "
            "(Internal use only)"
        ),
    )

    # We are currently inferring the site_id from the parent Attribute in
    # .save() method. We don't want to even care about the site_id, but it
    # simplifies managing them this way.
    site = models.ForeignKey(
        "Site",
        db_index=True,
        related_name="values",
        on_delete=models.PROTECT,
        verbose_name="Site",
        help_text="Unique ID of the Site this Value is under.",
    )

    def __init__(self, *args, **kwargs):
        self._obj = kwargs.pop("obj", None)
        super().__init__(*args, **kwargs)

    def __str__(self):
        return "%s:%s %s=%s" % (
            self.resource_name,
            self.resource_id,
            self.name,
            self.value,
        )

    class Meta:
        unique_together = ("name", "value", "resource_name", "resource_id")

        # This is most commonly looked up
        indexes = [
            models.Index(fields=["name", "value", "resource_name"]),
            models.Index(fields=["resource_name", "resource_id"]),
        ]

    def clean_resource_name(self, value):
        if value not in constants.VALID_ATTRIBUTE_RESOURCES:
            raise exc.ValidationError("Invalid resource name: %r." % value)
        return value

    def clean_name(self, attr):
        return attr.name

    def clean_site(self, value):
        """Always enforce that site is set."""
        if value is None:
            try:
                return self.attribute.site_id
            except Attribute.DoesNotExist:
                return Attribute.objects.get(id=self.attribute_id).site_id

        return value

    def clean_fields(self, exclude=None):
        obj = self._obj
        if obj is None:
            return

        self.site_id = self.clean_site(self.site_id)
        self.resource_name = self.clean_resource_name(obj.__class__.__name__)
        self.resource_id = obj.id
        self.name = self.clean_name(self.attribute)

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "value": self.value,
            "attribute": self.attribute_id,
            "resource_name": self.resource_name,
            "resource_id": self.resource_id,
        }



# -*- coding: utf-8 -*-

from django.db import models, migrations

def migrate_value_fields(apps, schema_editor):
    """
    Migrate new Value fields.
    """
    Value = apps.get_model("nsot", "Value")
    for val in Value.objects.iterator():
        val.resource_name = val.resource.polymorphic_ctype.model.title()
        val.new_resource_id = val.resource_id
        val.name = val.attribute.name
        val.save()

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0019_move_assignment_data"),
    ]

    operations = [
        # Value name, resource_id, resource_name
        migrations.RunPython(migrate_value_fields),
    ]



# Generated by Django 5.2.11 on 2026-02-12 23:03

import django_extensions.db.fields.json
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('nsot', '0041_interface_mtu'),
    ]

    operations = [
        migrations.AddField(
            model_name='attribute',
            name='_default',
            field=django_extensions.db.fields.json.JSONField(blank=True, db_column='default', default=None, help_text='Default value applied when this attribute is not provided on resource creation.', null=True, verbose_name='Default'),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations, models

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0032_add_indicies_to_change"),
    ]

    operations = [
        migrations.AddField(
            model_name="interface",
            name="name_slug",
            field=models.CharField(
                null=True,
                editable=False,
                max_length=255,
                help_text="Slugified version of the name field, used for the natural key",
                unique=True,
                db_index=True,
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations
import nsot.util
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
import nsot.fields
import django_extensions.db.fields.json

class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("auth", "0006_require_contenttypes_0002"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        null=True, verbose_name="last login", blank=True
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        unique=True,
                        max_length=255,
                        verbose_name="email address",
                        db_index=True,
                    ),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                (
                    "secret_key",
                    models.CharField(
                        default=nsot.util.generate_secret_key, max_length=44
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        related_query_name="user",
                        related_name="user_set",
                        to="auth.Group",
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        related_query_name="user",
                        related_name="user_set",
                        to="auth.Permission",
                        blank=True,
                        help_text="Specific permissions for this user.",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "verbose_name": "user",
                "verbose_name_plural": "users",
            },
        ),
        migrations.CreateModel(
            name="Attribute",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=64, db_index=True)),
                ("description", models.CharField(default="", max_length=255)),
                ("required", models.BooleanField(default=False)),
                ("display", models.BooleanField(default=False)),
                ("multi", models.BooleanField(default=False)),
                ("constraints", django_extensions.db.fields.json.JSONField(blank=True)),
                (
                    "resource_name",
                    models.CharField(
                        db_index=True,
                        max_length=20,
                        verbose_name="Resource Name",
                        choices=[("Network", "Network"), ("Device", "Device")],
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Change",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("change_at", models.DateTimeField(auto_now_add=True)),
                (
                    "event",
                    models.CharField(
                        max_length=10,
                        choices=[
                            ("Create", "Create"),
                            ("Update", "Update"),
                            ("Delete", "Delete"),
                        ],
                    ),
                ),
                ("resource_id", models.IntegerField(verbose_name="Resource ID")),
                (
                    "resource_name",
                    models.CharField(
                        db_index=True,
                        max_length=20,
                        verbose_name="Resource Type",
                        choices=[
                            ("Device", "Device"),
                            ("Attribute", "Attribute"),
                            ("Permission", "Permission"),
                            ("Site", "Site"),
                            ("Network", "Network"),
                        ],
                    ),
                ),
                (
                    "_resource",
                    django_extensions.db.fields.json.JSONField(
                        verbose_name="Resource", blank=True
                    ),
                ),
            ],
            options={
                "get_latest_by": "change_at",
            },
        ),
        migrations.CreateModel(
            name="Resource",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("_attributes", django_extensions.db.fields.json.JSONField(blank=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Site",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(unique=True, max_length=255)),
                ("description", models.TextField(default="", blank=True)),
            ],
        ),
        migrations.CreateModel(
            name="Value",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("value", models.CharField(db_index=True, max_length=255, blank=True)),
                (
                    "attribute",
                    models.ForeignKey(
                        related_name="values",
                        on_delete=django.db.models.deletion.PROTECT,
                        to="nsot.Attribute",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Device",
            fields=[
                (
                    "resource_ptr",
                    models.OneToOneField(
                        parent_link=True,
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        to="nsot.Resource",
                        on_delete=django.db.models.deletion.CASCADE,
                    ),
                ),
                ("hostname", models.CharField(max_length=255, db_index=True)),
                (
                    "site",
                    models.ForeignKey(
                        related_name="devices",
                        on_delete=django.db.models.deletion.PROTECT,
                        to="nsot.Site",
                    ),
                ),
            ],
            bases=("nsot.resource",),
        ),
        migrations.CreateModel(
            name="Network",
            fields=[
                (
                    "resource_ptr",
                    models.OneToOneField(
                        parent_link=True,
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        to="nsot.Resource",
                        on_delete=django.db.models.deletion.CASCADE,
                    ),
                ),
                (
                    "network_address",
                    nsot.fields.BinaryIPAddressField(max_length=16, db_index=True),
                ),
                (
                    "broadcast_address",
                    nsot.fields.BinaryIPAddressField(max_length=16, db_index=True),
                ),
                ("prefix_length", models.IntegerField(db_index=True)),
                (
                    "ip_version",
                    models.CharField(
                        db_index=True, max_length=1, choices=[("4", "4"), ("6", "6")]
                    ),
                ),
                ("is_ip", models.BooleanField(default=False, db_index=True)),
                (
                    "site",
                    models.ForeignKey(
                        related_name="networks",
                        on_delete=django.db.models.deletion.PROTECT,
                        to="nsot.Site",
                    ),
                ),
            ],
            bases=("nsot.resource",),
        ),
        migrations.AddField(
            model_name="value",
            name="resource",
            field=models.ForeignKey(
                related_name="attributes",
                blank=True,
                to="nsot.Resource",
                on_delete=django.db.models.deletion.CASCADE,
            ),
        ),
        migrations.AddField(
            model_name="resource",
            name="parent",
            field=models.ForeignKey(
                related_name="children",
                on_delete=django.db.models.deletion.PROTECT,
                default=None,
                blank=True,
                to="nsot.Resource",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="resource",
            name="polymorphic_ctype",
            field=models.ForeignKey(
                related_name="polymorphic_nsot.resource_set+",
                editable=False,
                to="contenttypes.ContentType",
                on_delete=django.db.models.deletion.CASCADE,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="change",
            name="site",
            field=models.ForeignKey(
                related_name="changes",
                to="nsot.Site",
                on_delete=django.db.models.deletion.CASCADE,
            ),
        ),
        migrations.AddField(
            model_name="change",
            name="user",
            field=models.ForeignKey(
                related_name="changes",
                to=settings.AUTH_USER_MODEL,
                on_delete=django.db.models.deletion.CASCADE,
            ),
        ),
        migrations.AddField(
            model_name="attribute",
            name="site",
            field=models.ForeignKey(
                related_name="attributes",
                on_delete=django.db.models.deletion.PROTECT,
                to="nsot.Site",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="value",
            unique_together=set([("attribute", "value", "resource")]),
        ),
        migrations.AlterIndexTogether(
            name="value",
            index_together=set([("attribute", "value", "resource")]),
        ),
        migrations.AlterUniqueTogether(
            name="attribute",
            unique_together=set([("site", "resource_name", "name")]),
        ),
        migrations.AlterIndexTogether(
            name="attribute",
            index_together=set([("site", "resource_name", "name")]),
        ),
        migrations.AlterUniqueTogether(
            name="network",
            unique_together=set(
                [("site", "ip_version", "network_address", "prefix_length")]
            ),
        ),
        migrations.AlterIndexTogether(
            name="network",
            index_together=set(
                [("site", "ip_version", "network_address", "prefix_length")]
            ),
        ),
        migrations.AlterUniqueTogether(
            name="device",
            unique_together=set([("site", "hostname")]),
        ),
        migrations.AlterIndexTogether(
            name="device",
            index_together=set([("site", "hostname")]),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations
import macaddress.fields

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0002_auto_20150810_1718"),
    ]

    operations = [
        migrations.AlterField(
            model_name="interface",
            name="mac",
            field=macaddress.fields.MACAddressField(
                db_index=True, integer=True, null=True, blank=True
            ),
        ),
    ]



import re

from django.conf import settings
from django.db import models

from .. import exc, fields, validators
from . import constants

# Internal key used to wrap default values for JSON serialization.
# The django-extensions JSONField doesn't properly serialize plain strings,
# so we wrap them in a dict with this key.
_DEFAULT_VALUE_KEY = "_v"


class Attribute(models.Model):
    """Represents a flexible attribute for Resource objects."""

    # This is purposely not unique as there is a compound index with site_id.
    name = models.CharField(
        max_length=64,
        null=False,
        db_index=True,
        help_text="The name of the Attribute.",
    )
    description = models.CharField(
        max_length=255,
        default="",
        blank=True,
        null=False,
        help_text="A helpful description of the Attribute.",
    )

    # The resource must contain a key and value
    required = models.BooleanField(
        default=False,
        null=False,
        help_text="Whether the Attribute should be required.",
    )

    # In UIs this attribute will be displayed by default. Required implies
    # display.
    display = models.BooleanField(
        default=False,
        null=False,
        help_text=(
            "Whether the Attribute should be be displayed by default in "
            "UIs. If required is set, this is also set."
        ),
    )

    # Attribute values are expected as lists of strings.
    multi = models.BooleanField(
        default=False,
        null=False,
        help_text="Whether the Attribute should be treated as a list type.",
    )

    constraints = fields.JSONField(
        "Constraints",
        null=False,
        blank=True,
        help_text="Dictionary of Attribute constraints.",
    )

    # Internal storage for default value. Use the `default` property instead.
    # Stored as {"_v": value} to ensure proper JSON serialization.
    _default = fields.JSONField(
        "Default",
        null=True,
        blank=True,
        default=None,
        db_column="default",
        help_text="Default value applied when this attribute is not provided on resource creation.",
    )

    @property
    def default(self):
        """Get the default value, unwrapping from internal storage format."""
        if self._default is None or self._default == {}:
            return None
        if (
            isinstance(self._default, dict)
            and _DEFAULT_VALUE_KEY in self._default
        ):
            return self._default[_DEFAULT_VALUE_KEY]
        # Fallback for any legacy data
        return self._default

    @default.setter
    def default(self, value):
        """Set the default value, wrapping for internal storage format."""
        if value is None:
            self._default = None
        else:
            self._default = {_DEFAULT_VALUE_KEY: value}

    depends_on = models.ManyToManyField(
        "self",
        symmetrical=False,
        blank=True,
        related_name="dependents",
        help_text="Attributes that must also be present on a resource when this Attribute is set.",
    )

    site = models.ForeignKey(
        "Site",
        db_index=True,
        related_name="attributes",
        on_delete=models.PROTECT,
        verbose_name="Site",
        help_text="Unique ID of the Site this Attribute is under.",
    )

    resource_name = models.CharField(
        "Resource Name",
        max_length=20,
        null=False,
        db_index=True,
        choices=constants.RESOURCE_CHOICES,
        help_text="The name of the Resource to which this Attribute is bound.",
    )

    def __str__(self):
        return "%s %s (site_id: %s)" % (
            self.resource_name,
            self.name,
            self.site_id,
        )

    class Meta:
        unique_together = ("site", "resource_name", "name")

    @classmethod
    def all_by_name(cls, resource_name=None, site=None):
        if resource_name is None:
            raise SyntaxError("You must provided a resource_name.")
        if site is None:
            raise SyntaxError("You must provided a site.")

        query = cls.objects.filter(resource_name=resource_name, site=site)

        return {attribute.name: attribute for attribute in query.all()}

    def clean_constraints(self, value):
        """Enforce formatting of constraints."""
        if not isinstance(value, dict):
            msg = f"Expected dictionary but received {type(value)}"
            raise exc.ValidationError({"constraints": msg})

        constraints = {
            "allow_empty": value.get("allow_empty", False),
            "pattern": value.get("pattern", ""),
            "valid_values": value.get("valid_values", []),
        }

        if not isinstance(constraints["allow_empty"], bool):
            raise exc.ValidationError(
                {"constraints": "allow_empty expected type bool."}
            )

        if not isinstance(constraints["pattern"], str):
            raise exc.ValidationError(
                {"constraints": "pattern expected type string."}
            )

        if not isinstance(constraints["valid_values"], list):
            raise exc.ValidationError(
                {"constraints": "valid_values expected type list."}
            )

        return constraints

    def clean_display(self, value):
        if self.required:
            return True
        return value

    def clean_resource_name(self, value):
        if value not in constants.VALID_ATTRIBUTE_RESOURCES:
            raise exc.ValidationError(
                {"resource_name": "Invalid resource name: %r." % value}
            )
        return value

    def clean_name(self, value):
        value = validators.validate_name(value)

        if not settings.ATTRIBUTE_NAME.match(value):
            raise exc.ValidationError(
                {
                    "name": "Invalid name: %r. Names must match: %s"
                    % (value, settings.ATTRIBUTE_NAME.pattern)
                }
            )

        return value or False

    def clean_default(self, value):
        """Validate the default value against multi type and constraints."""
        # No default set
        if value is None:
            return None

        # Validate type based on multi flag
        if self.multi:
            if not isinstance(value, list):
                raise exc.ValidationError(
                    {"default": "Default for multi attribute must be a list."}
                )
            # Ensure all items are strings
            if not all(isinstance(v, str) for v in value):
                raise exc.ValidationError(
                    {"default": "Default list items must be strings."}
                )
        else:
            if not isinstance(value, str):
                raise exc.ValidationError(
                    {
                        "default": "Default for single-value attribute must be a string."
                    }
                )

        # Validate the default value against constraints using validate_value()
        # This checks pattern, valid_values, and allow_empty
        try:
            self.validate_value(value)
        except exc.ValidationError as e:
            # Re-raise with "default" key for clarity
            raise exc.ValidationError({"default": str(e)})

        return value

    def clean_fields(self, exclude=None):
        self.constraints = self.clean_constraints(self.constraints)
        self.display = self.clean_display(self.display)
        self.resource_name = self.clean_resource_name(self.resource_name)
        self.name = self.clean_name(self.name)
        self.default = self.clean_default(self.default)

    def _validate_single_value(self, value, constraints=None):
        if not isinstance(value, str):
            raise exc.ValidationError(
                {"value": "Attribute values must be a string type"}
            )

        if constraints is None:
            constraints = self.constraints

        allow_empty = constraints.get("allow_empty", False)
        if not allow_empty and not value:
            msg = f"Attribute {self.name} doesn't allow empty values"
            raise exc.ValidationError({"constraints": msg})

        pattern = constraints.get("pattern")
        if pattern and not re.match(pattern, value):
            msg = f"Attribute value {value} for {self.name} didn't match pattern: {pattern}"
            raise exc.ValidationError({"pattern": msg})

        valid_values = set(constraints.get("valid_values", []))
        if valid_values and value not in valid_values:
            raise exc.ValidationError(
                "Attribute value {} for {} not a valid value: {}".format(
                    value, self.name, ", ".join(valid_values)
                )
            )

        return {
            "attribute_id": self.id,
            "value": value,
        }

    def validate_value(self, value):
        if self.multi:
            if not isinstance(value, list):
                raise exc.ValidationError(
                    {"multi": "Attribute values must be a list type"}
                )
        else:
            value = [value]

        inserts = []
        # This does a deserialization so save the result
        constraints = self.constraints
        for val in value:
            inserts.append(self._validate_single_value(val, constraints))

        return inserts

    def get_all_dependencies(self):
        """Return the full set of transitive dependencies (not just direct)."""
        visited = set()
        stack = list(self.depends_on.all())
        while stack:
            dep = stack.pop()
            if dep.pk not in visited:
                visited.add(dep.pk)
                stack.extend(dep.depends_on.all())
        return Attribute.objects.filter(pk__in=visited)

    def validate_dependencies(self):
        """Validate that depends_on entries are valid (same site/resource_name, no cycles)."""
        for dep in self.depends_on.all():
            if dep.pk == self.pk:
                raise exc.ValidationError(
                    {"depends_on": "An attribute cannot depend on itself."}
                )
            if dep.site_id != self.site_id:
                raise exc.ValidationError(
                    {
                        "depends_on": f"Dependency '{dep.name}' belongs to a different site."
                    }
                )
            if dep.resource_name != self.resource_name:
                raise exc.ValidationError(
                    {
                        "depends_on": f"Dependency '{dep.name}' is for resource type "
                        f"'{dep.resource_name}', expected '{self.resource_name}'."
                    }
                )

        # Cycle detection via DFS
        visited = set()
        stack = list(self.depends_on.all())
        while stack:
            dep = stack.pop()
            if dep.pk == self.pk:
                raise exc.ValidationError(
                    {"depends_on": "Circular dependency detected."}
                )
            if dep.pk not in visited:
                visited.add(dep.pk)
                stack.extend(dep.depends_on.all())

    def delete(self, *args, **kwargs):
        """Prevent deletion if other attributes depend on this one."""
        dependents = self.dependents.all()
        if dependents.exists():
            names = ", ".join(sorted(d.name for d in dependents))
            raise exc.ValidationError(
                {
                    "depends_on": f"Cannot delete attribute '{self.name}' because "
                    f"the following attributes depend on it: {names}"
                }
            )
        return super().delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        """Always enforce constraints."""
        self.full_clean()
        super().save(*args, **kwargs)

    def to_dict(self):
        return {
            "id": self.id,
            "site_id": self.site_id,
            "description": self.description,
            "name": self.name,
            "resource_name": self.resource_name,
            "required": self.required,
            "display": self.display,
            "multi": self.multi,
            "constraints": self.constraints,
            "default": self.default,
        }



# -*- coding: utf-8 -*-

from django.db import migrations, models

def remove_name_slug(apps, schema_editor):
    Interface = apps.get_model("nsot", "Interface")
    Interface.objects.update(name_slug=None)

class Migration(migrations.Migration):

    dependencies = [
        ("nsot", "0033_add_interface_name_slug"),
    ]

    operations = [
        # The "forwards" action was changed to a noop so that migration 0035
        # fully replaces this one without having to do migration gymnastics.
        # Users who previously upgraded to v1.2.0 and had migration 0034
        # already applied will have to apply 0035, but new users will have 0034
        # migration be a noop.
        # In summary:
        # - v1.2.0 =  0033 -> 0034 -> 0035
        # - v1.2.1 =  0033 -> 0035
        migrations.RunPython(migrations.RunPython.noop, remove_name_slug)
    ]



"""
Middleware to log HTTP requests.
"""

import logging
from time import time

from django.utils.deprecation import MiddlewareMixin

logger = logging.getLogger("nsot_server")


class LoggingMiddleware(MiddlewareMixin):
    def process_request(self, request):
        request.timer = time()
        return

    def process_response(self, request, response):
        if "HTTP_X_FORWARDED_FOR" in request.META:
            request_ip_path = "%s, %s" % (
                request.META.get("REMOTE_ADDR"),
                request.META.get("HTTP_X_FORWARDED_FOR"),
            )
        else:
            request_ip_path = request.META.get("REMOTE_ADDR")
        logger.info(
            "%s %s %s (%s) %.2fms",
            response.status_code,
            request.method,
            request.get_full_path(),
            request_ip_path,
            (time() - request.timer) * 1000,  # ms
        )
        return response



"""App configuration for nsot."""

from django.apps import AppConfig


class NsotConfig(AppConfig):
    """Default configuration for nsot."""

    name = "nsot"
    verbose_name = "Network Source of Truth"
